"""Represent collections of :class:`~SpectrumMatch` instances covering the same spectrum.

Also includes methods for selecting which are worth keeping for downstream consideration.
"""
import logging
from typing import Iterable, Iterator, List, Dict, Any, Optional, Type, Union, Sequence, TypeVar

from ms_deisotope.data_source import ProcessedScan

from glycresoft.chromatogram_tree import Unmodified
from glycresoft.task import log_handle

from .spectrum_match import SpectrumMatch, SpectrumReference, ScanWrapperBase, MultiScoreSpectrumMatch


logger = logging.getLogger(__name__)

MatchType = TypeVar('MatchType', bound=SpectrumMatch)


class SpectrumMatchRetentionStrategyBase(object):
    """
    A method for filtering :class:`SpectrumMatch` objects out of a list according to a specific criterion.

    Attributes
    ----------
    threshold: object
        Some abstract threshold
    """
    threshold: float

    def __init__(self, threshold):
        self.threshold = threshold

    def filter_matches(self, solution_set: 'SpectrumSolutionSet'):
        """
        Filter :class:`SpectrumMatch` objects from a list.

        Parameters
        ----------
        solution_set : list
            The list of :class:`SpectrumMatch` objects to filter.

        Returns
        -------
        list
        """
        raise NotImplementedError()

    def __call__(self, solution_set: 'SpectrumSolutionSet'):
        return self.filter_matches(solution_set)

    def __repr__(self):
        return "{self.__class__.__name__}({self.threshold})".format(self=self)


class MinimumScoreRetentionStrategy(SpectrumMatchRetentionStrategyBase):
    """
    Filtering :class:`~.SpectrumMatch` from a list if their :attr:`~.SpectrumMatch.score` < :attr:`threshold`.

    Parameters
    ----------
    solution_set : list
        The list of :class:`SpectrumMatch` objects to filter.

    Returns
    -------
    list
    """

    def filter_matches(self, solution_set):
        retain = []
        for match in solution_set:
            if match.score > self.threshold:
                retain.append(match)
        return retain


class MinimumMultiScoreRetentionStrategy(SpectrumMatchRetentionStrategyBase):
    """
    A minimum score filtering strategy for multi-score matches.

    A strategy for filtering :class:`~.SpectrumMatch` from a list if
    their :attr:`~.SpectrumMatch.score_set` is less than :attr:`threshold`,
    assuming they share the same dimensions.

    Parameters
    ----------
    solution_set : list
        The list of :class:`SpectrumMatch` objects to filter.

    Returns
    -------
    list
    """

    def filter_matches(self, solution_set):
        retain = []
        for match in solution_set:
            for score_i, ref_i in zip(match.score_set, self.threshold):
                if score_i < ref_i:
                    break
            else:
                retain.append(match)
        return retain


class MaximumSolutionCountRetentionStrategy(SpectrumMatchRetentionStrategyBase):
    """
    A strategy for filtering :class:`~.SpectrumMatch` from a list to retain the top :attr:`threshold` entries.

    This assumes that `solution_set` is sorted.

    Parameters
    ----------
    solution_set : list
        The list of :class:`SpectrumMatch` objects to filter.

    Returns
    -------
    list
    """

    def group_solutions(self, solutions, threshold=1e-2):
        """
        Group solutions which have scores that are very close to one-another so they are not arbitrarily truncated.

        Parameters
        ----------
        solutions : list
            A list of :class:`~.SpectrumMatchBase`
        threshold : float, optional
            The maxmimum distance between two scores to still be considered
            part of a group (the default is 1e-2)

        Returns
        -------
        list
        """
        groups = []
        if len(solutions) == 0:
            return groups
        current_group = [solutions[0]]
        last_solution = solutions[0]
        for solution in solutions[1:]:
            delta = abs(solution.score - last_solution.score)
            if delta > threshold:
                groups.append(current_group)
                current_group = [solution]
            else:
                current_group.append(solution)
            last_solution = solution
        groups.append(current_group)
        return groups

    def filter_matches(self, solution_set):
        groups = self.group_solutions(solution_set)
        return [b for a in groups[:self.threshold] for b in a]


class TopScoringSolutionsRetentionStrategy(SpectrumMatchRetentionStrategyBase):
    """
    Retain only solutions with :attr:`threshold` of the top score.

    A strategy for filtering :class:`~.SpectrumMatch` from a list to retain
    those with scores that are within :attr:`threshold` of the highest score in
    the set.

    This assumes that `solution_set` is sorted and that the highest score is at
    the 0th index..

    Parameters
    ----------
    solution_set : list
        The list of :class:`SpectrumMatch` objects to filter.

    Returns
    -------
    list
    """

    def filter_matches(self, solution_set):
        if len(solution_set) == 0:
            return solution_set
        best_score = solution_set[0].score
        retain = []
        for solution in solution_set:
            if (best_score - solution.score) < self.threshold:
                retain.append(solution)
        return retain


class QValueRetentionStrategy(SpectrumMatchRetentionStrategyBase):
    """
    Filtering :class:`~.SpectrumMatch` from a list if their :attr:`~.SpectrumMatch.q_value` < :attr:`threshold`.

    A strategy for filtering :class:`~.SpectrumMatch` from a list to retain
    those with q-value that are less than :attr:`threshold`.

    Parameters
    ----------
    solution_set : list
        The list of :class:`SpectrumMatch` objects to filter.

    Returns
    -------
    list
    """

    def filter_matches(self, solution_set):
        retain = []
        for match in solution_set:
            if match.q_value < self.threshold:
                retain.append(match)
        return retain


class SpectrumMatchRetentionMethod(SpectrumMatchRetentionStrategyBase):
    """
    Combine several :class:`SpectrumMatchRetentionStrategyBase` together.

    A collection of several :class:`SpectrumMatchRetentionStrategyBase`
    objects which are applied in order to iteratively filter out :class:`SpectrumMatch`
    objects.

    This class implements the same :class:`SpectrumMatchRetentionStrategyBase` API
    so it may be used interchangably with single strategies.

    Attributes
    ----------
    strategies: list
        The list of :class:`SpectrumMatchRetentionStrategyBase`
    """

    def __init__(self, strategies=None):  # pylint: disable=super-init-not-called
        if strategies is None:
            strategies = []
        self.strategies = strategies

    def filter_matches(self, solution_set):
        retained = list(solution_set)
        for strategy in self.strategies:
            retained = strategy(retained)
        return retained

    def __repr__(self):
        return "{self.__class__.__name__}({self.strategies!r})".format(self=self)


default_selection_method = SpectrumMatchRetentionMethod([
    MinimumScoreRetentionStrategy(4.),
    TopScoringSolutionsRetentionStrategy(3.),
    MaximumSolutionCountRetentionStrategy(100)
])


default_multiscore_selection_method = SpectrumMatchRetentionMethod([
    MinimumMultiScoreRetentionStrategy((1.0, 0., 0.)),
    TopScoringSolutionsRetentionStrategy(100.),
    MaximumSolutionCountRetentionStrategy(100),
])

class SpectrumMatchSortingStrategy(object):
    """A base strategy for sorting spectrum matches to rank the best solution."""

    def key_function(self, spectrum_match: SpectrumMatch):
        """
        Create the sorting key for the spectrum match.

        For consistency, the target's ``id`` attribute is used to order
        matches that have the same score so repeated searches with the same
        database produce the same ordering.

        Returns
        -------
        tuple
        """
        return (spectrum_match.valid, spectrum_match.score, spectrum_match.target.id)

    def sort(self, solution_set: 'SpectrumSolutionSet', maximize: bool = True) -> List[SpectrumMatch]:
        """
        Perform the sorting step.

        Parameters
        ----------
        solution_set: :class:`Iterable` of :class:`~.SpectrumMatch` instances
            The spectrum matches to sort
        maximize: bool
            Whether to sort ascending or descending

        Returns
        -------
        :class:`list` of :class:`~.SpectrumMatch` instances
        """
        return sorted(solution_set, key=self.key_function, reverse=maximize)

    def __call__(self, solution_set: 'SpectrumSolutionSet', maximize: bool = True) -> List[SpectrumMatch]:
        """
        Perform the sorting step.

        Parameters
        ----------
        solution_set: :class:`Iterable` of :class:`~.SpectrumMatch` instances
            The spectrum matches to sort
        maximize: bool
            Whether to sort ascending or descending

        Returns
        -------
        :class:`list` of :class:`~.SpectrumMatch` instances
        """
        return self.sort(solution_set, maximize=maximize)


class MultiScoreSpectrumMatchSortingStrategy(SpectrumMatchSortingStrategy):
    def key_function(self, spectrum_match: MultiScoreSpectrumMatch):
        return (spectrum_match.valid, spectrum_match.score_set, spectrum_match.target.id)


class NOParsimonyMixin(object):
    """
    Provides shared methods for a parsimony step re-ordering solutions when the top
    solution may be N-linked or O-linked.
    """

    threshold_percentile: float

    def __init__(self, threshold_percentile=0.5):
        self.threshold_percentile = threshold_percentile

    def get_score(self, solution):
        return solution.score_set.glycan_score

    def get_target(self, solution):
        return solution.target

    def hoist_equivalent_n_linked_solution(self, solution_set: Iterable[SpectrumMatch],
                                           maximize: bool=True) -> List[SpectrumMatch]:
        """
        Apply a parsimony step re-ordering solutions when the top solution may be N-linked or O-linked.

        The O-linked core motif is much simpler than the N-linked core, so it gets
        preferential treatment. This re-orders the matches so that the N-linked solution
        equivalent to the O-linked solution gets marked as the top solution if the
        N-linked solution's glycan score is within :attr:`threshold_percent` of the
        actual top scoring O-linked solution.

        Parameters
        ----------
        solution_set: :class:`Iterable` of :class:`~.SpectrumMatch` instances
            The spectrum matches to sort
        maximize: bool
            Whether to sort ascending or descending

        Returns
        -------
        :class:`list` of :class:`~.SpectrumMatch` instances
        """
        best_solution = solution_set[0]
        best_gp = self.get_target(best_solution)
        best_gc = best_gp.glycan_composition

        hoisted = []
        rest = [best_solution]
        if maximize:
            threshold = self.get_score(best_solution) * self.threshold_percentile
        else:
            threshold = self.get_score(best_solution) * (1 + (1 - self.threshold_percentile))
        i = 0
        for i, sm in enumerate(solution_set):
            if i == 0:
                continue

            if maximize:
                if self.get_score(sm) < threshold:
                    rest.append(sm)
                    break
            else:
                if self.get_score(sm) > threshold:
                    rest.append(sm)
                    break
            sm_target = self.get_target(sm)
            if (best_gp.base_sequence_equality(sm_target) and str(best_gc) == str(sm_target.glycan_composition)
                and sm_target.is_n_glycosylated()):
                hoisted.append(sm)
                log_handle.log(f"Hoisting {sm.target} for scan {sm.scan.scan_id!r}")
            else:
                rest.append(sm)
        rest.extend(solution_set[i + 1:])
        hoisted.extend(rest)
        return hoisted


class NOParsimonyMultiScoreSpectrumMatchSortingStrategy(NOParsimonyMixin, MultiScoreSpectrumMatchSortingStrategy):
    """
    A sorting strategy that applies prefers N-glycans to O-glycans on the same peptide.

    It applies parsimony to selecting the top solution when
    an N-linked solution is more parsimonious than an O-linked solution.
    """

    def sort(self, solution_set: Iterable[MultiScoreSpectrumMatch], maximize=True) -> List[MultiScoreSpectrumMatch]:
        solution_set = super(NOParsimonyMultiScoreSpectrumMatchSortingStrategy, self).sort(solution_set, maximize)
        if solution_set and solution_set[0].target.is_o_glycosylated():
            solution_set = self.hoist_equivalent_n_linked_solution(solution_set, maximize)
        return solution_set


single_score_sorter = SpectrumMatchSortingStrategy()
multi_score_sorter = MultiScoreSpectrumMatchSortingStrategy()
# This sort function isn't useful at this moment. The actual
# ordering is set in chromatogram_mapping
# multi_score_parsimony_sorter = NOParsimonyMultiScoreSpectrumMatchSortingStrategy(0.75)



class SpectrumSolutionSet(ScanWrapperBase, Sequence[MatchType]):
    """
    A collection of spectrum matches against a single scan with different structures.

    Implements the :class:`Sequence` interface.

    Attributes
    ----------
    scan: :class:`~.Scan`-like
        The matched scan
    solutions: list
        The distinct spectrum matches.
    score: float
        The best match's score
    """

    spectrum_match_type: Type[MatchType] = SpectrumMatch
    default_selection_method: SpectrumMatchRetentionMethod = default_selection_method

    scan: Union[ProcessedScan, SpectrumReference]
    solutions: List[MatchType]

    _target_map: Dict[Any, MatchType]
    _is_top_only: bool
    _is_sorted: bool
    _is_simplified: bool
    _q_value: Optional[float]

    def __init__(self, scan, solutions=None):
        if solutions is None:
            solutions = []
        self.scan = scan
        self.solutions = solutions
        self._is_sorted = False
        self._is_simplified = False
        self._is_top_only = False
        self._target_map = None
        self._q_value = None

    def is_ambiguous(self) -> bool:
        seen = set()

        for sm in self:
            if sm.is_best_match:
                seen.add(str(sm.target))
        return len(seen) > 1

    def is_multiscore(self) -> bool:
        """
        Check whether this match has been produced by summarizing a multi-score match, rather
        than a single score match.

        Returns
        -------
        bool
        """
        return False

    def _invalidate(self, invalidate_order: bool):
        if invalidate_order:
            self._is_sorted = False
        self._target_map = None
        self._q_value = None

    @property
    def score(self):
        """The best match's score.

        Returns
        -------
        float
        """
        return self.best_solution().score

    @property
    def q_value(self):
        """The best match's q-value.

        Returns
        -------
        float
        """
        if self._q_value is None:
            self._q_value = self.best_solution().q_value
        return self._q_value

    @q_value.setter
    def q_value(self, value):
        self._q_value = value

    def _make_target_map(self):
        self._target_map = {
            sol.target: sol for sol in self
        }

    def has_solution(self, target) -> bool:
        if self._target_map is None:
            self._make_target_map()
        return target in self._target_map

    def solution_for(self, target) -> MatchType:
        """
        Find the spectrum match from this set which corresponds to the provided `target` structure.

        Parameters
        ----------
        target : object
            The target to search for

        Returns
        -------
        :class:`~.SpectrumMatchBase`
        """
        if self._target_map is None:
            self._make_target_map()
        return self._target_map[target]

    def precursor_mass_accuracy(self):
        """
        The precursor mass accuracy of the best match.

        Returns
        -------
        float
        """
        return self.best_solution().precursor_mass_accuracy()

    def best_solution(self, reject_shifted=False, targets_ignored=None, require_valid=True) -> MatchType:
        """
        The element in :attr:`solutions` which is the best match to :attr:`scan`, the match at position 0.

        If the collection is not sorted, :meth:`sort` will be called.

        Parameters
        ----------
        reject_shifted : bool
            Whether or not to reject any solution where the mass shift is not :obj:`Unmodified`.
            Defaults to :const:`False`.
        targets_ignored : Container
            A collection of :attr:`~.SpectrumMatch.target` values to ignore

        Returns
        -------
        :class:`~.SpectrumMatchBase`
        """
        if not self._is_sorted:
            self.sort()
        if targets_ignored is None:
            targets_ignored = ()
        if not reject_shifted and not targets_ignored:
            for solution in self.solutions:
                if solution.valid:
                    return solution
        for solution in self:
            if (solution.mass_shift == Unmodified or not reject_shifted) and (solution.target in targets_ignored or
                not targets_ignored) and (solution.valid or not require_valid):
                return solution

    def mark_top_solutions(self, reject_shifted=False, targets_ignored=None) -> 'SpectrumSolutionSet':
        solution = self.best_solution(
            reject_shifted=reject_shifted,
            targets_ignored=targets_ignored)
        if solution is None and reject_shifted:
            return self.mark_top_solutions(reject_shifted=False, targets_ignored=targets_ignored)
        if solution is None and targets_ignored:
            return self.mark_top_solutions(
                reject_shifted=reject_shifted, targets_ignored=None)
        if solution is None and not reject_shifted and not targets_ignored:
            logger.warn(f"Could not mark a top solution for {self.scan_id} (not reject shifted and not targets ignore)")
            return self
        if solution is None:
            logger.warn(f"Could not mark a top solution for {self.scan_id}")
            return self
        solution.best_match = True
        best_solution_score = solution.score
        for solution in self:
            if (abs(best_solution_score - solution.score) < 1e-3 or
                best_solution_score < solution.score) and solution.valid:
                solution.best_match = True
            else:
                solution.best_match = False
        return self

    def __repr__(self):
        if len(self) == 0:
            return "{self.__class__.__name__}({self.scan}, [])".format(self=self)
        return "{self.__class__.__name__}({self.scan}, {best.target}, {best.score})".format(
            self=self, best=self.best_solution())

    def __getitem__(self, i):
        return self.solutions[i]

    def __iter__(self) -> Iterator[MatchType]:
        return iter(self.solutions)

    def __len__(self):
        return len(self.solutions)

    def simplify(self):
        """
        Discard excess information in this collection to save space.

        Converts :attr:`scan` to a :class:`~.SpectrumReference`, and
        converts all matches to :class:`~.SpectrumMatch`, discarding
        matcher-specific information.

        """
        if self._is_simplified:
            return
        self.scan = SpectrumReference(
            self.scan.id, self.scan.precursor_information)
        solutions = []
        if len(self) > 0:
            best_score = self.best_solution().score
            for sol in self.solutions:
                sm = self.spectrum_match_type.from_match_solution(sol)
                if abs(sm.score - best_score) < 1e-6:
                    sm.best_match = True
                sm.scan = self.scan
                solutions.append(sm)
            self.solutions = solutions
        self._is_simplified = True
        self._invalidate(False)

    def get_top_solutions(self, d=3, reject_shifted=False, targets_ignored=None,
                          require_valid=True) -> List[MatchType]:
        """
        Get all matches within `d` of the best solution.

        Parameters
        ----------
        d : float, optional
            The delta between the best match and the worst to return
            (the default is 3)
        reject_shifted : bool
            Whether or not to reject any solution where the mass shift is not :obj:`Unmodified`.
            Defaults to :const:`False`.
        targets_ignored : Container
            A collection of :attr:`~.SpectrumMatch.target` values to ignore

        Returns
        -------
        list
        """
        best = self.best_solution(
            reject_shifted=reject_shifted, targets_ignored=targets_ignored, require_valid=require_valid)
        if best is None and reject_shifted:
            return self.get_top_solutions(
                d,
                reject_shifted=False,
                targets_ignored=targets_ignored,
                require_valid=require_valid
            )
        if best is None and targets_ignored:
            return self.get_top_solutions(
                d,
                reject_shifted=reject_shifted,
                targets_ignored=None,
                require_valid=require_valid
            )
        if best is None:
            return []
        best_score = best.score
        if reject_shifted:
            return [x for x in self.solutions if (best_score - x.score) < d and x.mass_shift == Unmodified]
        return [x for x in self.solutions if (best_score - x.score) < d]

    def select_top(self, method=None) -> 'SpectrumSolutionSet':
        """
        Filter spectrum matches in this collection in-place.

        If all the solutions would be filtered out, only the best solution
        will be kept.

        .. warning::
            This method is dangerous to use when spectrum matches in the same
            list are part of different groups (e.g. different decoy sets). Use
            it with caution.

        Parameters
        ----------
        method : :class:`SpectrumRetentionStrategyBase`, optional
            The filtering strategy to use, a callable object that returns a
            shortened list of :class:`~.SpectrumMatchBase` instances.
            If :const:`None`, :attr:`default_selection_method` will be used.

        """
        if method is None:
            method = self.default_selection_method
        if self._is_top_only:
            return self
        if not self._is_sorted:
            self.sort()
        if len(self) > 0:
            best_solution = self.best_solution()
            after = method(self)
            self.solutions = after
            if len(self) == 0:
                self.solutions = [best_solution]
        self._is_top_only = True
        self._invalidate(False)
        return self

    def sort(self, maximize=True, method=None) -> 'SpectrumSolutionSet':
        """
        Sort the spectrum matches in this solution set according to their score attribute.

        In the event of a tie, in order to enforce determistic behavior, this will also
        sort matches according to their target's id attribute.

        Sets :attr:`_is_sorted` to :const:`True`.

        Parameters
        ----------
        maximize : bool, optional
            If true, sort descending order instead of ascending. Defaults to :const:`True`

        See Also
        --------
        sort_by
        sort_q_value
        """
        if method is None:
            method = single_score_sorter
        self.solutions = method(self.solutions, maximize=maximize)
        self._is_sorted = True
        return self

    def sort_by(self, sort_fn=None, maximize=True) -> 'SpectrumSolutionSet':
        """
        Sort the spectrum matches in this solution set according to `sort_fn`.

        This method behaves the same way :meth:`sort` does, except instead of
        sorting on an intrinsic attribute it uses a callable. It uses the same
        determistic augmentation as :meth:`sort`

        Parameters
        ----------
        sort_fn : Callable, optional
            The sort key function to use. If not provided, falls back to :meth:`sort`.
        maximize : bool, optional
            If true, sort descending order instead of ascending. Defaults to :const:`True`

        See Also
        --------
        sort
        """
        if sort_fn is None:
            return self.sort(maximize=maximize)
        self.solutions.sort(key=lambda x: (sort_fn(x), x.target.id), reverse=maximize)
        self._is_sorted = True
        return self

    def sort_q_value(self) -> 'SpectrumSolutionSet':
        """
        Sort the spectrum matches in this solution set according to their q_value attribute.

        In the event of a tie, in order to enforce determistic behavior, this will also
        sort matches according to their target's id attribute.

        Sets :attr:`_is_sorted` to :const:`True`.

        See Also
        --------
        sort
        sort_by
        """
        self.solutions.sort(key=lambda x: (x.q_value, x.target.id), reverse=False)
        self._is_sorted = True
        return self

    def merge(self, other) -> 'SpectrumSolutionSet':
        self._invalidate(True)
        self.solutions.extend(other)
        self.sort()
        if self._is_top_only:
            self._is_top_only = False
            self.select_top()
        return self

    def append(self, match: MatchType) -> 'SpectrumSolutionSet':
        self._invalidate(True)
        self.solutions.append(match)
        self.sort()
        if self._is_top_only:
            self._is_top_only = False
            self.select_top()
        return self

    def insert(self, position: int, match: MatchType, is_sorted=True):
        was_sorted = self._is_sorted
        self._invalidate(True)
        if is_sorted:
            self._is_sorted = was_sorted
        self.solutions.insert(position, match)

    def remove(self, match: MatchType):
        self.solutions.remove(match)

    def promote_to_best_match(self, match: MatchType, reject_shifted: bool=False,
                              targets_ignored: Optional[List[Any]] = None) -> 'SpectrumSolutionSet':
        try:
            self.remove(match)
        except ValueError:
            pass
        self.insert(0, match, is_sorted=True)
        return self.mark_top_solutions(
            reject_shifted=reject_shifted,
            targets_ignored=targets_ignored
        )

    def clone(self):
        dup = self.__class__(self.scan, [
            s.clone() for s in self.solutions
        ])
        dup._is_simplified = self._is_simplified
        dup._is_top_only = self._is_top_only
        dup._is_sorted = self._is_sorted
        return dup

    def __eq__(self, other):
        if self.scan.id != other.scan.id:
            return False
        return self.solutions == other.solutions

    def __ne__(self, other):
        return not (self == other)

    @property
    def key(self) -> frozenset:
        scan_id = self.scan.id
        return frozenset([(scan_id, match.target.id) for match in self])

    def rank(self) -> 'SpectrumSolutionSet':
        """
        Rank the spectrum matches in this solution set by their ordering and scores.

        The best match receiving the lowest rank and successive matches receiving
        higher ranks, the best rank being 1. Spectrum matches with scores that are
        very close together will receive the same rank.

        The score ranking process is only approximate because in some scenarios the
        match ordering is overridden by external factors e.g. retention time, signature
        ions, or adducts.

        Returns
        -------
        SpectrumSolutionSet
        """
        if not self:
            return self
        current_rank = 0
        current_score = float('inf')
        for sm in self:
            if abs(sm.score - current_score) >= 1e-3:
                current_rank += 1
                current_score = sm.score
            sm.rank = current_rank
        return self


def close_to_or_greater_than(value: float, reference: float, delta: float=1e-3):
    return abs(value - reference) < delta or value > reference


class MultiScoreSpectrumSolutionSet(SpectrumSolutionSet[MultiScoreSpectrumMatch]):
    spectrum_match_type = MultiScoreSpectrumMatch
    default_selection_method = default_multiscore_selection_method

    def is_multiscore(self):
        return True

    # NOTE: Sorting by total score is not guaranteed to sort by total
    # FDR. Sorting by FDR after-the-fact is cheating though because we
    # don't count next-best decoys.

    def sort(self, maximize=True, method=None):
        """
        Sort the spectrum matches in this solution set according to their score_set attribute.

        In the event of a tie, in order to enforce determistic behavior, this will also
        sort matches according to their target's id attribute.

        Sets :attr:`_is_sorted` to :const:`True`.

        See Also
        --------
        sort_by
        sort_q_value
        """
        if method is None:
            method = multi_score_sorter
        self.solutions = method(self.solutions, maximize=maximize)
        self._is_sorted = True
        return self

    def mark_top_solutions(self, reject_shifted=False, targets_ignored=None):
        solution = self.best_solution(reject_shifted=reject_shifted, targets_ignored=targets_ignored)
        if solution is None and reject_shifted:
            return self.mark_top_solutions(reject_shifted=False, targets_ignored=targets_ignored)
        if solution is None and targets_ignored:
            return self.mark_top_solutions(reject_shifted=reject_shifted, targets_ignored=None)
        if solution is None and not reject_shifted and not targets_ignored:
            logger.warn(f"Could not mark a top solution for {self.scan_id}")
            return self
        solution.best_match = True
        score_set = solution.score_set
        for solution in self:
            if not solution.valid:
                solution.best_match = False
                continue
            if reject_shifted and solution.mass_shift != Unmodified:
                solution.best_match = False
                continue
            if targets_ignored and solution.target in targets_ignored:
                solution.best_match = False
                continue
            if ((score_set.glycopeptide_score - solution.score_set.glycopeptide_score)) < 1e-3:
                # NOTE: This requires that both the peptide portion and the glycan portion be as
                # good or better than the "best" match. When the "best match" is better in one
                # dimension and worse in another dimension, this becomes more ambiguous. It's
                # possible to query the FDR of the cases, but :meth:`mark_top_solutions` may
                # be called before FDR estimation. Furthermore, a "best match" ranked by FDR
                # potentially introduces all sorts of issues of re-ranking, and could introduce
                # an undesirable number of "alternative" best matches.
                if (score_set.peptide_score - solution.score_set.peptide_score) < 1e-3 and (
                    score_set.glycan_score - solution.score_set.glycan_score) < 1e-3:
                    solution.best_match = True
                else:
                    solution.best_match = False
            else:
                solution.best_match = False
        return self

    def sort_q_value(self):
        """
        Sort the spectrum matches in this solution set according to their q_value_set attribute.

        In the event of a tie, in order to enforce determistic behavior, this will also
        sort matches according to their target's id attribute.

        Sets :attr:`_is_sorted` to :const:`True`.

        See Also
        --------
        sort
        sort_by
        """
        self.solutions.sort(key=lambda x: (
            x.q_value_set, x.score_set, x.target.id), reverse=False)
        self._is_sorted = True
        return self

    @property
    def score_set(self):
        """
        The best match's score set.

        Returns
        -------
        ScoreSet
        """
        return self.best_solution().score_set

    @property
    def q_value_set(self):
        """
        The best match's q-value set.

        Returns
        -------
        FDRSet
        """
        return self.best_solution().q_value_set



from glycresoft.database.builder.glycan import TextFileGlycanHypothesisSerializer
from glycresoft.database.prebuilt.utils import hypothesis_register, BuildBase
from io import StringIO


hypothesis_metadata = {
    "name": "Glycosaminoglycan Linkers",
    "hypothesis_type": "glycan_composition",
    "description": ''
}

source_text = u'''{Xyl:1; a,enHex:1; Hex:2; aHex:1; HexNAc:1}      GAG-linker
{Xyl:1; a,enHex:1; Hex:2; aHex:1; HexNAc(S):1}   GAG-linker
{Xyl:1; Fuc:1; a,enHex:1; Hex:2; aHex:1; HexNAc:1}       GAG-linker
{Xyl:1; Fuc:1; a,enHex:1; Hex:1; aHex:1; HexS:1; HexNAc(S):1}    GAG-linker
{Xyl:1; Fuc:1; a,enHex:1; Hex:2; aHex:1; HexNAc(S):1}    GAG-linker
{Fuc:1; a,enHex:1; Hex:2; aHex:1; HexNAc:1; XylP:1}      GAG-linker
{a,enHex:1; Hex:2; aHex:1; HexNAc:1; XylP:1}     GAG-linker
{a,enHex:1; Hex:2; aHex:1; XylP:1; HexNAc(S):1}  GAG-linker
{a,enHex:1; Hex:1; aHex:1; XylP:1; HexS:1; HexNAc(S):1}  GAG-linker
{Xyl:1; a,enHex:1; Hex:2; aHex:1; HexNAc(S):1}   GAG-linker
{Xyl:1; a,enHex:1; Hex:1; aHex:1; HexS:1; HexNAc(S):1}   GAG-linker
{Xyl:1; a,enHex:1; aHex:1; HexS:2; HexNAc(S):1}  GAG-linker
{Xyl:1; a,enHex:1; Hex:2; aHex:1; HexNAc:1; Neu5Ac:1}    GAG-linker
{Xyl:1; a,enHex:1; Hex:2; aHex:1; HexNAc(S):1; Neu5Ac:1}         GAG-linker
{Xyl:1; a,enHex:1; Hex:1; aHex:1; HexS:1; HexNAc(S):1; Neu5Ac:1}         GAG-linker
{Xyl:1; a,enHex:1; Hex:2; aHex:1; HexNAc:1; Neu5Ac:1}    GAG-linker
{Xyl:1; a,enHex:1; Hex:2; aHex:1; HexNAc(S):1; Neu5Ac:1}         GAG-linker
{a,enHex:1; Hex:2; aHex:1; XylP:1; HexNAc(S):1; Neu5Ac:1}        GAG-linker
{Xyl:1; a,enHex:1; Hex:1; aHex:1; HexS:1; HexNAc(S):1; Neu5Ac:1}         GAG-linker
{a,enHex:1; Hex:1; aHex:1; XylP:1; HexS:1; HexNAc(S):1; Neu5Ac:1}        GAG-linker
{Xyl:1; a,enHex:1; Hex:2; aHex:1; HexNAc:1; Neu5Gc:1}    GAG-linker
{Xyl:1; a,enHex:1; Hex:2; aHex:1; HexNAc(S):1; Neu5Gc:1}         GAG-linker
{Xyl:1; a,enHex:1; Hex:1; aHex:1; HexS:1; HexNAc(S):1; Neu5Gc:1}         GAG-linker
{Xyl:1; a,enHex:1; Hex:2; aHex:1; HexNAc:1; Neu5Gc:1}    GAG-linker
{Xyl:1; a,enHex:1; Hex:2; aHex:1; HexNAc(S):1; Neu5Gc:1}         GAG-linker
{a,enHex:1; Hex:2; aHex:1; XylP:1; HexNAc(S):1; Neu5Gc:1}        GAG-linker
{Xyl:1; a,enHex:1; Hex:1; aHex:1; HexS:1; HexNAc(S):1; Neu5Gc:1}         GAG-linker
{a,enHex:1; Hex:1; aHex:1; XylP:1; HexS:1; HexNAc(S):1; Neu5Gc:1}        GAG-linker'''


@hypothesis_register(hypothesis_metadata['name'])
class GlycosaminoglycanLinkersBuilder(BuildBase):

    def get_hypothesis_metadata(self):
        return hypothesis_metadata

    def prepare_buffer(self):
        text_buffer = StringIO(source_text)
        return text_buffer

    def build(self, database_connection, **kwargs):
        kwargs.setdefault('hypothesis_name', self.hypothesis_metadata['name'])
        task = TextFileGlycanHypothesisSerializer(
            self.prepare_buffer(), database_connection,
            **kwargs)
        task.start()
        return task



from matplotlib import pyplot as plt
import numpy as np


def target_decoy_score_separation(target_hits, decoy_hits, binwidth=5, ax=None):
    if ax is None:
        fig, ax = plt.subplots(1)
    bins = np.arange(min([s.score for s in target_hits]), max([s.score for s in target_hits]) + binwidth, binwidth)
    ax.hist([s.score for s in target_hits], bins=bins, alpha=0.4, lw=0.5, ec='white', label='Target PSMs')
    ax.hist([s.score for s in decoy_hits], bins=bins, alpha=0.4, lw=0.5, ec='white', label='Decoy PSMs')
    ax.legend().get_frame().set_linewidth(0)
    return ax



# -*- coding: utf-8 -*-
import itertools
import math
import array

from collections import defaultdict
from decimal import Decimal
from typing import Any, DefaultDict, Dict, List, Optional, Set, Tuple, Union, NamedTuple, Deque

import numpy as np
from scipy.special import comb

from glycopeptidepy import PeptideSequence, Modification, ModificationRule, IonSeries, PeptideFragment
from glycopeptidepy.utils.memoize import memoize
from glycopeptidepy.algorithm import PeptidoformGenerator, ModificationSiteAssignmentCombinator

from ms_deisotope.data_source import ProcessedScan
from ms_deisotope.peak_set import window_peak_set, DeconvolutedPeak
from ms_deisotope.peak_dependency_network.intervals import SpanningMixin

from glycresoft.structure import FragmentCachingGlycopeptide
from glycresoft.structure.probability import PredictorBase

from glycresoft.tandem.spectrum_match.spectrum_match import LocalizationScore, ScanMatchManagingMixin


MAX_MISSING_A_SCORE = 1e3


@memoize(100000000000)
def binomial_pmf(n, i, p):
    try:
        return comb(n, i, exact=True) * (p ** i) * ((1 - p) ** (n - i))
    except OverflowError:
        dn = Decimal(n)
        di = Decimal(i)
        dp = Decimal(p)
        x = math.factorial(dn) / (math.factorial(di) * math.factorial(dn - di))
        return float(x * dp ** di * ((1 - dp) ** (dn - di)))


class PeakWindow(SpanningMixin):
    peaks: List[DeconvolutedPeak]
    max_mass: float

    def __init__(self, peaks):
        self.peaks = list(peaks)
        self.max_mass = 0
        self.start = 0
        self.end = 0
        self._calculate()

    def __iter__(self):
        return iter(self.peaks)

    def __getitem__(self, i):
        return self.peaks[i]

    def __len__(self):
        return len(self.peaks)

    def _calculate(self):
        self.peaks.sort(key=lambda x: x.intensity, reverse=True)
        max_mass = 0
        min_mass = float('inf')
        for peak in self.peaks:
            if peak.neutral_mass > max_mass:
                max_mass = peak.neutral_mass
            elif peak.neutral_mass < min_mass:
                min_mass = peak.neutral_mass
        self.end = self.max_mass = max_mass
        self.start = min_mass

    def __repr__(self):
        template = "{self.__class__.__name__}({self.max_mass}, {size})"
        return template.format(self=self, size=len(self))



class BlindPeptidoformGenerator(PeptidoformGenerator):
    """A sub-class of the :class:`~.PeptidoformGenerator` type that ignores
    site-specificities.
    """
    def modification_sites_for(self, sequence, rules: List[ModificationRule], include_empty: bool=True):
        variable_sites = {
            mod: set(range(len(sequence))) for mod in rules}
        modification_sites = ModificationSiteAssignmentCombinator(
            variable_sites, include_empty=include_empty)
        return modification_sites


class ModificationAssignment(NamedTuple):
    site: int
    modification: Modification

    @property
    def is_ambiguous(self):
        try:
            return len(self.site) > 1
        except TypeError:
            return False

    def itersites(self):
        if self.is_ambiguous:
            for i in self.site:
                yield i
        else:
            yield self.site


class ProbableSitePair(NamedTuple):
    top_solution1: PeptideSequence
    top_solution2: PeptideSequence
    modifications: List[Modification]
    peak_depth: int


class AScoreSpec(NamedTuple):
    site: ModificationAssignment
    score: float
    site_determining_ions: np.array
    alternative_site_ions: np.array


class LocalizationCandidate(object):
    peptide: PeptideSequence
    modifications: List[Union[ModificationRule, Modification]]
    fragments: List[PeptideFragment]

    def __init__(self, peptide, modifications, fragments=None):
        self.peptide = peptide
        self.modifications = modifications
        self.fragments = fragments

    def __hash__(self):
        return hash(self.peptide)

    def __eq__(self, other):
        return self.peptide == other.peptide and self.modifications == other.modifications

    def make_solution(self, a_score: AScoreSpec, permutations: Optional[np.ndarray] = None) -> 'AScoreSolution':
        return AScoreSolution(self.peptide, a_score, self.modifications, permutations, self.fragments)

    def __repr__(self):
        template = "{self.__class__.__name__}({d})"

        def formatvalue(v):
            if isinstance(v, float):
                return "%0.4f" % v
            else:
                return str(v)
        d = [
            "%s=%s" % (k, formatvalue(v)) if v is not self else "(...)" for k, v in sorted(
                self.__dict__.items(), key=lambda x: x[0])
            if (not k.startswith("_") and not callable(v))
            and not (v is None) and k != "fragments"]

        return template.format(self=self, d=', '.join(d))


class AScoreSolution(LocalizationCandidate):
    a_score: List[AScoreSpec]
    permutations: np.ndarray

    def __init__(self, peptide: PeptideSequence, a_score, modifications: List[Modification], permutations,
                 fragments: Optional[List[PeptideFragment]]=None):
        super().__init__(peptide, modifications, fragments)
        self.a_score = a_score
        self.permutations = permutations


class PeptidoformPermuter(object):
    peptide: PeptideSequence
    modification_rule: ModificationRule
    modification_count: int
    respect_specificity: bool

    def __init__(self, peptide: PeptideSequence, modification_rule: ModificationRule,
                 modification_count: int=1, respect_specificity: bool=True):
        self.peptide = peptide
        self.modification_rule = modification_rule
        self.modification_count = modification_count
        self.respect_specificity = respect_specificity

    def find_existing(self, modification_rule: ModificationRule):
        '''Find existing modifications derived from this rule

        Parameters
        ----------
        modification_rule: :class:`~.ModificationRule`
            The modification rule to search for

        Returns
        -------
        indices: list
            The indices of :attr:`peptide` where modifications were found
        '''
        indices = []
        for i, position in enumerate(self.peptide):
            if position.modifications and modification_rule in position.modifications:
                indices.append(i)
        return indices

    def generate_base_peptides(self, modification_rule: ModificationRule) -> List[PeptideSequence]:
        """Generate peptides from :attr:`peptide` which have had combinations of
        modification sites removed.

        Parameters
        ----------
        modification_rule : :class:`~.ModificationRule`
            The modification rule to remove

        Returns
        -------
        list
        """
        existing_indices = self.find_existing(modification_rule)
        base_peptides: List[Union[PeptideSequence, FragmentCachingGlycopeptide]] = []
        for indices in itertools.combinations(existing_indices, self.modification_count):
            base_peptide = self.peptide.clone()
            for i in indices:
                base_peptide.drop_modification(i, modification_rule)
            base_peptides.append(base_peptide)
        # The target modification was not present, so the unaltered peptide must be the base
        if not base_peptides:
            base_peptides = [self.peptide.clone()]
        if isinstance(base_peptides[0], FragmentCachingGlycopeptide):
            for bp in base_peptides:
                bp.clear_caches()
        return base_peptides

    def generate_peptidoforms(self, modification_rule: ModificationRule, base_peptides: Optional[PeptideSequence]=None):
        if base_peptides is None:
            base_peptides = self.generate_base_peptides(modification_rule)
        if self.respect_specificity:
            PeptidoformGeneratorType = PeptidoformGenerator
        else:
            PeptidoformGeneratorType = BlindPeptidoformGenerator
        pepgen = PeptidoformGeneratorType(
            [], [modification_rule], self.modification_count)
        peptidoforms = defaultdict(set)
        for base_peptide in base_peptides:
            is_caching = isinstance(base_peptide, FragmentCachingGlycopeptide)
            mod_combos = pepgen.modification_sites_for(base_peptide, pepgen.variable_modifications)
            for mod_combo in mod_combos:
                if len(mod_combo) != self.modification_count:
                    continue
                mod_combo = [ModificationAssignment(*mc) for mc in mod_combo]
                peptidoform, _n_mods = pepgen.apply_variable_modifications(
                    base_peptide, mod_combo, None, None)
                if is_caching:
                    peptidoform.clear_caches()
                peptidoforms[peptidoform].update(tuple(mod_combo))
        return [LocalizationCandidate(peptide, sorted(mods), self._generate_fragments(peptide))
                for peptide, mods in peptidoforms.items()]


class LocalizationScorerBase(PeptidoformPermuter, ScanMatchManagingMixin):

    scan: ProcessedScan
    _fragment_cache: Dict[Union[PeptideSequence, Any], List[PeptideFragment]]

    def __init__(self, scan, peptide: PeptideSequence, modification_rule: ModificationRule,
                 modification_count: int = 1, respect_specificity: bool = True):
        super().__init__(peptide, modification_rule, modification_count, respect_specificity)
        self.scan = scan
        self._fragment_cache = {}

    def _generate_fragments(self, peptidoform: PeptideSequence) -> List[PeptideFragment]:
        key = str(peptidoform)
        if key in self._fragment_cache:
            return self._fragment_cache[key]

        ion_series = []

        if self.is_hcd():
            ion_series.append(IonSeries.b)
            ion_series.append(IonSeries.y)
        if self.is_exd():
            ion_series.append(IonSeries.c)
            ion_series.append(IonSeries.z)

        frags = itertools.chain.from_iterable(
            itertools.chain.from_iterable(
                map(peptidoform.get_fragments, ion_series)
            )
        )
        frags = sorted(frags, key=lambda x: x.mass)
        self._fragment_cache[key] = frags
        return frags

    def clear_cache(self):
        self._fragment_cache.clear()


class AScoreEvaluator(LocalizationScorerBase):
    '''
    Calculate a localization statistic for given peptidoform and modification rule.

    The original probabilistic model is described in [1]. Implementation based heavily
    on the OpenMS implementation [2].

    References
    ----------
    [1] Beausoleil, S. a, Villén, J., Gerber, S. a, Rush, J., & Gygi, S. P. (2006).
        A probability-based approach for high-throughput protein phosphorylation analysis
        and site localization. Nature Biotechnology, 24(10), 1285–1292. https://doi.org/10.1038/nbt1240
    [2] Rost, H. L., Sachsenberg, T., Aiche, S., Bielow, C., Weisser, H., Aicheler, F., … Kohlbacher, O. (2016).
        OpenMS: a flexible open-source software platform for mass spectrometry data analysis. Nat Meth, 13(9),
        741–748. https://doi.org/10.1038/nmeth.3959
    '''
    peak_windows: List[PeakWindow]

    def __init__(self, scan, peptide, modification_rule, modification_count=1, respect_specificity=True):
        self._scan = None
        super().__init__(
            scan, peptide, modification_rule, modification_count, respect_specificity)
        self.peptidoforms = self.generate_peptidoforms(self.modification_rule)

    @property
    def scan(self):
        return self._scan

    @scan.setter
    def scan(self, value):
        self._scan = value
        if value is None:
            self.peak_windows = []
        else:
            self.peak_windows = list(map(PeakWindow, window_peak_set(value.deconvoluted_peak_set)))


    def match_ions(self, fragments: List[PeptideFragment], depth: int=10, error_tolerance: float=1e-5) -> Tuple[int, List[float]]:
        '''Match fragments against the windowed peak set at a given
        peak depth.

        Parameters
        ----------
        fragments: list
            A list of peptide fragments, sorted by mass
        depth: int
            The peak depth to search to, the `i`th most intense peak in
            each window
        error_tolerance: float
            The PPM error tolerance to use when matching peaks.

        Returns
        -------
        int:
            The number of fragments matched
        '''
        n = 0
        window_i = 0
        window_n = len(self.peak_windows)
        current_window = self.peak_windows[window_i]
        intensities: array.ArrayType[float] = array.array('f')
        for frag in fragments:
            while not current_window or (frag.mass >= (current_window.max_mass + 1)):
                window_i += 1
                if window_i == window_n:
                    return n
                current_window = self.peak_windows[window_i]
            for peak in current_window[:depth]:
                if abs(peak.neutral_mass - frag.mass) / frag.mass < error_tolerance:
                    intensities.append(peak.intensity)
                    n += 1
        return n, np.frombuffer(intensities, dtype=np.float32)

    def permutation_score(self, peptidoform: LocalizationCandidate, error_tolerance: float=1e-5) -> Tuple[np.ndarray,
                                                                                                    List[List[float]]]:
        '''Calculate the binomial statistic for this peptidoform
        using the top 1 to 10 peaks.

        Parameters
        ----------
        peptidoform: :class:`~.PeptideSequence`
            The peptidoform to score
        error_tolerance: float
            The PPM error tolerance to use when matching peaks.

        Returns
        -------
        :class:`numpy.ndarray`:
            The binomial score at peak depth `i + 1`

        See Also
        --------
        :meth:`_score_at_window_depth`
        :meth:`match_ions`
        '''
        fragments = peptidoform.fragments
        N = len(fragments)
        site_scores = np.zeros(10)
        intensities = [None for i in range(10)]
        for i in range(1, 11):
            site_scores[i - 1], intensities[i - 1] = self._score_at_window_depth(
                fragments, N, i, error_tolerance)
        return site_scores, intensities

    def _score_at_window_depth(self, fragments: List[PeptideFragment], N: int, i: int, error_tolerance: float=1e-5) -> Tuple[float, List[float]]:
        '''Score a fragment collection at a given peak depth, and
        calculate the binomial score based upon the probability mass
        function.

        Parameters
        ----------
        fragments: list
            A list of peptide fragments, sorted by mass
        N: int
            The maximum number of theoretical fragments
        i: int
            The peak depth to search through
        error_tolerance: float
            The PPM error tolerance to use when matching peaks.

        Returns
        -------
        score: float
        intensities: list[float]
        '''
        n, intensities = self.match_ions(fragments, i, error_tolerance=error_tolerance)
        p = i / 100.0
        # If a fragment matches twice, this count can exceed the theoretical maximum.
        if n > N:
            n = N
        cumulative_score = binomial_pmf(N, n, p)
        if cumulative_score == 0.0:
            return MAX_MISSING_A_SCORE, intensities
        return abs(-10.0 * math.log10(cumulative_score)), intensities

    def rank_permutations(self, permutation_scores):
        """Rank generated peptidoforms by weighted sum of permutation scores

        Parameters
        ----------
        permutation_scores : :class:`list` of :class:`list` of :class:`float`
            The raw output of :meth:`permutation_score` for each peak depth for
            each peptidoform.

        Returns
        -------
        :class:`list`
            A list of :class:`tuple` instances of (weighted score, peptidoform index)
        """
        ranking = []
        for i, perm_scores in enumerate(permutation_scores):
            ranking.append((self._weighted_score(perm_scores), i))
        ranking.sort(reverse=True)
        return ranking

    # Taken directly from reference [1]
    _weight_vector = np.array([
        0.5, 0.75, 1.0, 1.0, 1.0, 1.0, 0.75, 0.5, .25, .25
    ])

    def _weighted_score(self, scores):
        """Calculate the weighted sum score over the peak-depth permuted
        binomial score vector.

        Parameters
        ----------
        scores : :class:`list`
            The binomial score at each peak depth

        Returns
        -------
        float
        """
        return self._weight_vector.dot(scores) / 10.0

    def score_solutions(self, error_tolerance: float=1e-5, peptidoforms: Optional[List[LocalizationCandidate]]=None) -> List[AScoreSolution]:
        if peptidoforms is None:
            peptidoforms = self.peptidoforms
        scores, _intensities = zip(*[self.permutation_score(candidate, error_tolerance=error_tolerance)
                  for candidate in peptidoforms])
        ranked = self.rank_permutations(scores)
        solutions = [peptidoforms[i].make_solution(score, scores[i])
                     for score, i in ranked]
        return solutions

    def score_localizations(self, solutions: List[AScoreSolution], error_tolerance=1e-5):
        """Find pairs of sequence solutions which differ in the localization
        of individual modifications w.r.t. to the best match to compute the final
        per-modification A-score.

        The first solution in `solutions` is the highest ranked solution, and subsequent
        solutions are searched for the next case where one of the modification of interest
        is located at a different position, forming a pair for that modification site by
        :meth:`find_highest_scoring_permutations`. For each pair, the sequences are re-scored
        using only site-determining ions, and the difference between those scores is the A-score
        for that pair's modification site, as calculated by :meth:`calculate_delta`.

        If there are no alternative sites for a given modification, that modification will be
        given the A-score given by :const:`MAX_MISSING_A_SCORE`. If there is another
        localization which scores equally well, the A-score will be 0 by definition of
        the delta step.

        Parameters
        ----------
        solutions : list
            The list of :class:`AScoreSolution` objects, ranked by total score
        error_tolerance : float, optional
            The mass error tolerance to use when matching site-determining ions (the default is 1e-5)

        Returns
        -------
        :class:`AScoreSolution`
        """
        delta_scores = []
        pairs = self.find_highest_scoring_permutations(solutions)
        top_solution = solutions[0]
        if not pairs:
            for mod in top_solution.modifications:
                delta_scores.append(
                    AScoreSpec(
                        mod,
                        MAX_MISSING_A_SCORE,
                        np.array([], dtype=np.float32),
                        np.array([], dtype=np.float32)
                    )
                )
            top_solution.a_score = delta_scores
            return top_solution
        for pair in pairs:
            delta_score, intensities1, intensities2 = self.calculate_delta(pair, error_tolerance=error_tolerance)
            pair.top_solution1.a_score = delta_score
            delta_scores.append(
                AScoreSpec(
                    pair.modifications,
                    delta_score,
                    intensities1,
                    intensities2
                )
            )
        top_solution.a_score = delta_scores
        return top_solution

    def score(self, error_tolerance: float=1e-5) -> AScoreSolution:
        solutions = self.score_solutions(error_tolerance)
        top_solution = self.score_localizations(solutions, error_tolerance)
        return top_solution

    def find_highest_scoring_permutations(self, solutions: List[AScoreSolution],
                                          best_solution: Optional[AScoreSolution]=None,
                                          offset: Optional[int] = None) -> List[ProbableSitePair]:
        if best_solution is None:
            best_solution = solutions[0]
            offset = 1
        else:
            if offset is None:
                for i, sol in enumerate(solutions):
                    if sol == solutions:
                        offset = i + 1
                        break
                else:
                    raise ValueError("Best solution %r not in solution set")
        permutation_pairs = []
        # for each modification under permutation, find the next best solution which
        # does not have this modification in its set of permuted modifications, and
        # package the pair into a :class:`ProbableSitePair`.
        for site in best_solution.modifications:
            for alt_solution in solutions[offset:]:
                if site not in alt_solution.modifications:
                    peak_depth = np.argmax(best_solution.permutations - alt_solution.permutations) + 1
                    permutation_pairs.append(ProbableSitePair(best_solution, alt_solution, site, peak_depth))
                    break
        return permutation_pairs

    def site_determining_ions(self, solutions: List[LocalizationCandidate]) -> List[List[PeptideFragment]]:
        frag_sets = [set(sol.fragments) for sol in solutions]
        counts = defaultdict(int)
        for frag_set in frag_sets:
            for frag in frag_set:
                counts[frag] += 1
        site_determining = []
        for frag_set in frag_sets:
            acc = []
            for frag in frag_set:
                if counts[frag] == 1:
                    acc.append(frag)
            site_determining.append(sorted(acc, key=lambda x: x.mass))
        return site_determining

    def calculate_delta(self, candidate_pair: ProbableSitePair, error_tolerance=1e-5):
        if candidate_pair.top_solution1 == candidate_pair.top_solution2:
            return 0.0
        site_frags = self.site_determining_ions(
            [candidate_pair.top_solution1, candidate_pair.top_solution2])
        site_frags1, site_frags2 = site_frags[0], site_frags[1]
        N1 = len(site_frags1)
        N2 = len(site_frags2)
        peak_depth = candidate_pair.peak_depth
        score1, intensities1 = self._score_at_window_depth(
            site_frags1, N1, peak_depth, error_tolerance=error_tolerance)
        score2, intensities2 = self._score_at_window_depth(
            site_frags2, N2, peak_depth, error_tolerance=error_tolerance)
        return score1 - score2, intensities1, intensities2


class _PeakSet:
    peaks: Dict[int, float]
    total: float

    def __init__(self, peaks=None, total=None):
        if not peaks:
            peaks = {}
            total = 0
        self.peaks = peaks
        self.total = total

    def add(self, peak: DeconvolutedPeak):
        self.peaks[peak.index.neutral_mass] = peak.intensity
        self.total += peak.intensity

    def shared_intensity(self, other: '_PeakSet'):
        common = self.peaks.keys() & other.peaks.keys()
        acc = 0
        for k in common:
            acc += self.peaks[k]
        return acc

    def count_partitions(self, other: '_PeakSet'):
        common = self.peaks.keys() & other.peaks.keys()
        m = len(common)
        return len(self.peaks) - m, len(other.peaks) - m, m



class _PTMProphetMatch(NamedTuple):
    occupied: _PeakSet
    unoccupied: _PeakSet
    normalizer: float

    def o_score(self) -> float:
        common = self.occupied.shared_intensity(self.unoccupied)
        occupied_diff = (self.occupied.total - common) / self.normalizer
        unoccupied_diff = (self.unoccupied.total - common) / self.normalizer
        return occupied_diff / (occupied_diff + unoccupied_diff + 1e-6)

    def m_score(self) -> float:
        m_occupied, m_unoccupied, _shared = self.occupied.count_partitions(self.unoccupied)
        return m_occupied / (m_occupied + m_unoccupied + 1e-6)

    def to_solution(self, occupied_index: int, unoccupied_index: int) -> 'PTMProphetSolution':
        return PTMProphetSolution(self.o_score(), self.m_score(), occupied_index, unoccupied_index)


class PTMProphetSolution(NamedTuple):
    o_score: float
    m_score: float
    occupied_peptidoform_index: int
    unoccupied_peptidoform_index: int


class ScoredIsoform(NamedTuple):
    isoform: LocalizationCandidate
    localizations: List[LocalizationScore]
    score: float


class PTMProphetEvaluator(LocalizationScorerBase):
    raw_matches: List[_PeakSet]
    occupied_site_index: DefaultDict[int, Set[int]]
    solution_for_site: Dict[int, PTMProphetSolution]

    def __init__(self, scan, peptide, modification_rule, modification_count=1, respect_specificity=True):
        super().__init__(
            scan, peptide, modification_rule, modification_count, respect_specificity)
        self.peptidoforms = self.generate_peptidoforms(self.modification_rule)
        self.raw_matches = []
        self.occupied_site_index = DefaultDict(set)
        self.solution_for_site = {}

    def get_normalizer(self) -> float:
        normalizer = float('inf')
        for peak in self.scan.deconvoluted_peak_set:
            if peak.intensity < normalizer:
                normalizer = peak.intensity
        return normalizer

    def score_arrangements(self, error_tolerance: float=1e-5):
        raw_matches = []
        occupied_site_index = DefaultDict[int, Set[int]](set)

        normalizer = self.get_normalizer()

        for i, peptidoform in enumerate(self.peptidoforms):
            acc = _PeakSet()
            for fragment in peptidoform.fragments:
                for peak in self.scan.deconvoluted_peak_set.all_peaks_for(fragment.mass, error_tolerance):
                    acc.add(peak)
            raw_matches.append(acc)
            for j, pos in enumerate(peptidoform.peptide):
                if pos.has_modification(self.modification_rule):
                    occupied_site_index[j].add(i)

        self.raw_matches = raw_matches
        self.occupied_site_index = occupied_site_index

        for site, occupied_in in self.occupied_site_index.items():
            best_occupied_match = _PeakSet()
            best_occupied_index = None
            best_unoccupied_match = _PeakSet()
            best_unoccupied_index = None
            for i in range(len(self.peptidoforms)):
                if i in occupied_in:
                    if best_occupied_match.total < self.raw_matches[i].total:
                        best_occupied_match = self.raw_matches[i]
                        best_occupied_index = i
                else:
                    if best_unoccupied_match.total < self.raw_matches[i].total:
                        best_unoccupied_match = self.raw_matches[i]
                        best_unoccupied_index = i

            if best_occupied_index is None:
                continue
            self.solution_for_site[site] = _PTMProphetMatch(
                best_occupied_match, best_unoccupied_match, normalizer).to_solution(
                    best_occupied_index, best_unoccupied_index)

    def __repr__(self):
        return f"{self.__class__.__name__}({self.scan}, {self.peptide}, {self.modification_rule}, {self.solution_for_site})"

    def entropy(self, prophet: Optional[PredictorBase]) -> float:
        m = self.modification_count
        s = len(self.solution_for_site)
        log_base = np.log(s / m)
        acc = 0.0
        for sol in self.solution_for_site.values():
            prob = prophet.predict([sol.o_score, sol.m_score])[0]
            acc += prob * np.log(prob) / log_base
        return acc

    def site_probabilities(self, prophet: Optional[PredictorBase]) -> Dict[int, float]:
        position_probs = {}
        acc = 0.0
        for position, sol in self.solution_for_site.items():
            if prophet is not None:
                prob = prophet.predict([sol.o_score, sol.m_score])[0]
            else:
                prob = sol.o_score
            position_probs[position] = prob
            acc += prob
        acc /= self.modification_count
        factor = 1 / acc if acc > 0 else 0
        return {k: min(v * factor, 1.0) for k, v in position_probs.items()}

    def score_isoforms(self, prophet: Optional[PredictorBase]) -> List[ScoredIsoform]:
        weights: List[float] = []
        localizations: List[List[LocalizationScore]] = []
        sites = self.site_probabilities(prophet)

        for candidate in self.peptidoforms:
            acc = 0.0
            parts: List[LocalizationScore] = []
            for mod_a in candidate.modifications:
                try:
                    score = sites[mod_a.site]
                except KeyError:
                    if not isinstance(mod_a.site, int):
                        # TODO: This needs to handle terminal modifications
                        # which requires a change to the core algorithm.
                        # Ignore them for now.
                        score = 0.0
                    else:
                        # The position was never observed. Ignore it.
                        # May happen when sites are not restricted during
                        # peptidoform generation here but not during the original
                        # search.
                        score = 0.0
                acc += score
                parts.append(
                    LocalizationScore(
                        mod_a.site,
                        mod_a.modification.name,
                        score
                    )
                )
            localizations.append(parts)
            weights.append(acc)

        scored_isoforms = []
        for candidate, weight, loc_scores in zip(self.peptidoforms, weights, localizations):
            scored_isoforms.append(ScoredIsoform(
                candidate, loc_scores, weight))
        return scored_isoforms



from .base import (
    ScoreCell,
    ScoreThresholdCounter,
    ArrayScoreThresholdCounter,
    NearestValueLookUp,
    TargetDecoyAnalyzer,
    TargetDecoySet,
    GroupwiseTargetDecoyAnalyzer,
    PeptideScoreTargetDecoyAnalyzer,
    FDREstimatorBase,
)

from .svm import (
    SVMModelBase,
    PeptideScoreSVMModel
)

__all__ = [
    "ScoreCell",
    "ScoreThresholdCounter",
    "ArrayScoreThresholdCounter",
    "NearestValueLookUp",
    "TargetDecoyAnalyzer",
    "TargetDecoySet",
    "GroupwiseTargetDecoyAnalyzer",
    "PeptideScoreTargetDecoyAnalyzer",
    "SVMModelBase",
    "PeptideScoreSVMModel",
    "FDREstimatorBase",
]



import json


def read_colors_from_file(path):
    with open(path, 'rt') as fh:
        data = json.load(fh)
    data = {
        k: tuple(v) for k, v in data.items()
    }
    return data


def write_colors_to_file(colors, path):
    with open(path, 'wt') as fh:
        json.dump(colors, fh, sort_keys=True, indent=2)



from glycresoft.scoring.base import (
    CompositionDispatchingModel,
    is_sialylated,
    DummyFeature)

from glycresoft.scoring import (
    MassScalingChargeStateScoringModel, ChromatogramScorer)

from .utils import make_model_loader
from .features import register_feature


load_model = make_model_loader(MassScalingChargeStateScoringModel)


SialylatedChargeModel = load_model("sialylated_charge_model")
UnsialylatedChargeModel = load_model("unsialylated_charge_model")

GeneralizedChargeScoringModel = CompositionDispatchingModel({
    is_sialylated: SialylatedChargeModel,
    lambda x: not is_sialylated(x): UnsialylatedChargeModel
}, SialylatedChargeModel)


GeneralScorer = ChromatogramScorer()
GeneralScorer.add_feature(GeneralizedChargeScoringModel)


register_feature("null_charge", DummyFeature("null_charge_model", "charge_count"))



class SpectrumReference(object):
    def __init__(self, scan_id, precursor_information=None):
        self.scan_id = scan_id
        self.precursor_information = precursor_information

    @property
    def id(self):
        return self.scan_id

    def __eq__(self, other):
        try:
            return (self.scan_id == other.scan_id) and (
                self.precursor_information == other.precursor_information)
        except AttributeError:
            return self.scan_id == str(other)

    def __str__(self):
        return str(self.scan_id)

    def __hash__(self):
        return hash(self.scan_id)

    def __repr__(self):
        return "SpectrumReference({self.id})".format(self=self)


class TargetReference(object):
    def __init__(self, id):
        self.id = id

    def __repr__(self):
        return "TargetReference({self.id})".format(self=self)



from glypy.composition import formula as _formulastr, Composition
from ms_deisotope.utils import dict_proxy
from collections import defaultdict


@dict_proxy("composition")
class Formula(object):
    """Represent a hashable composition wrapper and a backing object supplying
    other attributes and properties. Used to make handling of objects which posses
    composition-like behavior easier.

    This class implements the Mapping interface, supplying its composition as key-value data.

    Attributes
    ----------
    composition : Composition
        The elemental composition to proxy
    data : object
        An arbitrary object to proxy attribute look ups to.
        Usually the source of :attr:`composition`
    name : str
        The formula for the given composition. Used for hashing and equality testing
    """
    def __init__(self, composition, data):
        self.composition = composition
        self.name = _formulastr(composition)
        self.data = data
        self._hash = hash(self.name)

    def __getattr__(self, attr):
        if attr == "data":
            return None
        return getattr(self.data, attr)

    def __eq__(self, other):
        return self.name == other.name

    def __hash__(self):
        return self._hash

    def total_composition(self):
        return Composition(self.composition)


@dict_proxy("composition")
class CommonComposition(object):
    """A type to represent multiple objects which share the same elemental composition,
    holding a list of all the objects which it shares formulae with.

    This class implements the Mapping interface, supplying its composition as key-value data.

    Attributes
    ----------
    base_data : list
        All the objects which are bundled together.
    composition : Composition
        The composition being represented.
    """
    def __init__(self, composition, base_data):
        self.composition = composition
        self.base_data = base_data

    @property
    def name(self):
        return '|'.join(x.name for x in self.base_data)

    @classmethod
    def aggregate(cls, iterable):
        """Combine an iterable of objects which are hashable and function as
        composition mappings into a list of :class:`CommonComposition` instances such
        that each composition is represented exactly once.

        Parameters
        ----------
        iterable : Iterable
            An iterable of any type of object which is both hashable and provides a Composition-like
            interface

        Returns
        -------
        list
        """
        agg = defaultdict(list)
        for item in iterable:
            agg[item].append(item.data)
        results = []
        for key, collected in agg.items():
            results.append(cls(key, collected))
        return results

    def __repr__(self):
        return "CommonComposition(%s)" % dict(self.composition)

    def total_composition(self):
        return Composition(self.composition)



from .lru import LRUCache, LRUNode, LRUMapping
from .structure_loader import (
    CachingGlycanCompositionParser,
    CachingGlycopeptideParser,
    FragmentCachingGlycopeptide,
    PeptideProteinRelation,
    DecoyFragmentCachingGlycopeptide,
    SequenceReversingCachingGlycopeptideParser,
    GlycopeptideCache,
    CachingPeptideParser,
    PeptideDatabaseRecord,)

from .scan import (
    ScanStub,
    ScanWrapperBase,
    ScanInformation,
    ScanInformationLoader)
from .fragment_match_map import FragmentMatchMap, SpectrumGraph
from .utils import KeyTransformingDecoratorDict


__all__ = [
    "LRUNode",
    "LRUCache",
    "LRUMapping",
    "CachingGlycanCompositionParser",
    "CachingGlycopeptideParser",
    "CachingPeptideParser",
    "FragmentCachingGlycopeptide",
    "PeptideProteinRelation",
    "DecoyFragmentCachingGlycopeptide",
    "SequenceReversingCachingGlycopeptideParser",
    "GlycopeptideCache",
    "ScanStub",
    "ScanWrapperBase",
    "ScanInformation",
    "ScanInformationLoader",
    "FragmentMatchMap",
    "SpectrumGraph",
    "KeyTransformingDecoratorDict",
    "PeptideDatabaseRecord",
]



from .binomial_score import BinomialSpectrumMatcher
from glycresoft.tandem import target_decoy
from glycresoft.tandem.target_decoy import (TargetDecoyAnalyzer, GroupwiseTargetDecoyAnalyzer)
from .precursor_mass_accuracy import MassAccuracyModel, MassAccuracyMixin
from .simple_score import SimpleCoverageScorer
from .coverage_weighted_binomial import (
    CoverageWeightedBinomialScorer, ShortPeptideCoverageWeightedBinomialScorer,
    CoverageWeightedBinomialModelTree)
from .intensity_scorer import (LogIntensityScorer, LogIntensityModelTree,
                               HyperscoreScorer, FullSignaturePenalizedLogIntensityScorer)
from .base import SpectrumMatcherBase, GlycopeptideSpectrumMatcherBase


__all__ = [
    "BinomialSpectrumMatcher", "target_decoy", "TargetDecoyAnalyzer", "GroupwiseTargetDecoyAnalyzer",
    "MassAccuracyModel", "MassAccuracyMixin", "SimpleCoverageScorer", "CoverageWeightedBinomialScorer",
    "ShortPeptideCoverageWeightedBinomialScorer", "CoverageWeightedBinomialModelTree",
    "GlycopeptideSpectrumMatcherBase", "SpectrumMatcherBase",
    "LogIntensityScorer", "LogIntensityModelTree", "HyperscoreScorer", "FullSignaturePenalizedLogIntensityScorer"
]



from collections import defaultdict

from typing import Any, Dict, Generic, List, Tuple, TypeVar

import numpy as np

from ms_deisotope.data_source import ProcessedScan


T = TypeVar('T')


class MassWrapper(Generic[T]):
    obj: T
    mass: float

    def __init__(self, obj, mass):
        self.obj = obj
        self.mass = mass

    def __repr__(self):
        return "MassWrapper(%s, %f)" % (self.obj, self.mass)


class NodeBase(object):
    def __init__(self, index: int):
        self.index = index
        self._hash = hash(self.index)
        self.edges = EdgeSet()

    def __hash__(self):
        return self._hash

    def __eq__(self, other):
        return self.index == other.index

    def has_incoming_edge(self):
        for edge in self.edges:
            if edge.is_incoming(self):
                return True
        return False


class EdgeSet(object):
    store: Dict[Tuple[NodeBase, NodeBase], 'GraphEdgeBase']

    def __init__(self, store=None):
        if store is None:
            store = dict()
        self.store = store

    def edge_to(self, node1: NodeBase, node2: NodeBase):
        if node2.index < node1.index:
            node1, node2 = node2, node1
        return self.store.get((node1, node2))

    def add(self, edge: 'GraphEdgeBase'):
        self.store[edge.node1, edge.node2] = edge

    def remove(self, edge: 'GraphEdgeBase'):
        self.store.pop((edge.node1, edge.node2))

    def __len__(self):
        return len(self.store)

    def __iter__(self):
        return iter(self.store.values())

    def __repr__(self):
        template = '{self.__class__.__name__}({self.store})'
        return template.format(self=self)


class AnnotatedMultiEdgeSet(EdgeSet):
    def __init__(self, store=None):
        if store is None:
            store = defaultdict(dict)
        EdgeSet.__init__(self, store)

    def __iter__(self):
        for subgroup in self.store.values():
            for edge in subgroup.values():
                yield edge

    def add(self, edge):
        self.store[edge.node1, edge.node2][edge.annotation] = edge

    def __len__(self):
        return sum(map(len, self.store.values()))

    def __repr__(self):
        template = '{self.__class__.__name__}({self.store})'
        return template.format(self=self)


class GraphEdgeBase(object):
    __slots__ = ["node1", "node2", "_hash", "_str", "indices"]

    node1: NodeBase
    node2: NodeBase
    indices: Tuple[int, int]

    def __init__(self, node1, node2):
        self.node1 = node1
        self.node2 = node2

        self.indices = (self.node1.index, self.node2.index)

        self._hash = self._make_hash()
        self._str = self._make_str()

        node1.edges.add(self)
        node2.edges.add(self)

    def _make_hash(self):
        return hash((self.node1, self.node2))

    def _make_str(self):
        return "GraphEdge(%s)" % ', '.join(map(str, (self.node1, self.node2)))

    def __getitem__(self, key: NodeBase):
        if key == self.node1:
            return self.node2
        elif key == self.node2:
            return self.node1
        else:
            raise KeyError(key)

    def is_incoming(self, node: NodeBase):
        if node is self.node2:
            return True
        elif node is self.node1:
            return False
        else:
            return False

    def _traverse(self, node):
        return self.node1 if node is self.node2 else self.node2

    def __eq__(self, other):
        return self._str == other._str

    def __str__(self):
        return self._str

    def __repr__(self):
        return self._str

    def __hash__(self):
        return self._hash

    def __reduce__(self):
        return self.__class__, (self.node1, self.node2,)

    def copy_for(self, node1: NodeBase, node2: NodeBase):
        return self.__class__(node1, node2,)

    def remove(self):
        try:
            self.node1.edges.remove(self)
        except KeyError:
            pass
        try:
            self.node2.edges.remove(self)
        except KeyError:
            pass


NodeT = TypeVar('NodeT', bound=NodeBase)

class GraphBase(Generic[NodeT]):
    edges: EdgeSet
    nodes: List[NodeT]

    def __init__(self, edges=None):
        if edges is None:
            edges = EdgeSet()
        self.edges = edges
        self.nodes = []

    def __iter__(self):
        return iter(self.nodes)

    def __getitem__(self, i):
        node = self.nodes[i]
        return node

    def __len__(self):
        return len(self.nodes)

    def adjacency_matrix(self):
        N = len(self)
        A = np.zeros((N, N), dtype=np.uint8)
        for edge in self.edges:
            A[edge.indices] = 1
        return A

    def topological_sort(self, adjacency_matrix: np.ndarray = None) -> List[NodeT]:
        if adjacency_matrix is None:
            adjacency_matrix = self.adjacency_matrix()
        else:
            adjacency_matrix = adjacency_matrix.copy()
        waiting = set()
        for i in range(adjacency_matrix.shape[0]):
            # Check for incoming edges. If no incoming
            # edges, add to the waiting set
            if adjacency_matrix[:, i].sum() == 0:
                waiting.add(i)

        ordered = list()
        while waiting:
            ix = waiting.pop()
            node = self[ix]
            ordered.append(node)
            # Get the set of outgoing edge terminal indices
            outgoing_edges = adjacency_matrix[ix, :]
            indices_of_neighbors = np.nonzero(outgoing_edges)[0]
            # For each outgoing edge
            for neighbor_ix in indices_of_neighbors:
                # Remove the edge
                adjacency_matrix[ix, neighbor_ix] = 0
                # Test for incoming edges
                if (adjacency_matrix[:, neighbor_ix] == 0).all():
                    waiting.add(neighbor_ix)
        if adjacency_matrix.sum() > 0:
            raise ValueError("%d edges left over" % (adjacency_matrix.sum(),))
        else:
            return ordered

    def path_lengths(self):
        adjacency_matrix = self.adjacency_matrix()
        distances = np.zeros(len(self))
        for node in self.topological_sort():
            incoming_edges = np.nonzero(adjacency_matrix[:, node.index])[0]
            if incoming_edges.shape[0] == 0:
                distances[node.index] = 0
            else:
                longest_path_so_far = distances[incoming_edges].max()
                distances[node.index] = longest_path_so_far + 1
        return distances

    def connected_components(self) -> List[List[NodeT]]:
        pool = set(self.nodes)
        components = []

        i = 0
        while pool:
            i += 1
            current_component = set()
            visited = set()
            node = pool.pop()
            current_component.add(node)
            j = 0
            while current_component:
                j += 1
                node = current_component.pop()
                visited.add(node)
                for edge in node.edges:
                    if edge.node1 not in visited and edge.node1 in pool:
                        current_component.add(edge.node1)
                        pool.remove(edge.node1)
                    if edge.node2 not in visited and edge.node2 in pool:
                        current_component.add(edge.node2)
                        pool.remove(edge.node2)
            components.append(list(visited))
        return components


class ScanNode(NodeBase):
    scan: ProcessedScan

    def __init__(self, scan, index):
        self.scan = scan
        NodeBase.__init__(self, index)
        self.edges = AnnotatedMultiEdgeSet()

    @property
    def id(self):
        return self.scan.id

    @property
    def precursor_information(self):
        return self.scan.precursor_information

    def __repr__(self):
        return "ScanNode(%s)" % (self.scan.id,)


class ScanGraphEdge(GraphEdgeBase):
    __slots__ = ["annotation"]

    def __init__(self, node1, node2, annotation):
        self.annotation = annotation
        GraphEdgeBase.__init__(self, node1, node2)

    def _make_hash(self):
        return hash((self.node1, self.node2, self.annotation))

    def _make_str(self):
        return "ScanGraphEdge(%s)" % ', '.join(map(str, (self.node1, self.node2, self.annotation)))

    def __reduce__(self):
        return self.__class__, (self.node1, self.node2, self.annotation)

    def copy_for(self, node1, node2):
        return self.__class__(node1, node2, self.annotation)


class ScanGraph(GraphBase[ScanNode]):
    scans: List[ProcessedScan]
    node_map: Dict[str, ScanNode]

    def __init__(self, scans):
        GraphBase.__init__(self)
        self.scans = scans
        self.nodes = [ScanNode(scan, i) for i, scan in enumerate(scans)]
        self.node_map = {node.scan.id: node for node in self.nodes}
        self.edges = AnnotatedMultiEdgeSet()

    def get_node_by_id(self, scan_id: str):
        return self.node_map[scan_id]

    def add_edge_between(self, scan_id1: str, scan_id2: str, annotation=None):
        node1 = self.get_node_by_id(scan_id1)
        node2 = self.get_node_by_id(scan_id2)
        edge = ScanGraphEdge(node1, node2, annotation)
        self.edges.add(edge)



from typing import Optional, Sequence, Iterable, TypeVar

from glycresoft.task import TaskBase

from .spectrum_match.solution_set import SpectrumSolutionSet
from .target_decoy import TargetDecoySet


T = TypeVar('T')


def chunkiter(collection: Sequence[T], size: int=200) -> Iterable[Sequence[T]]:
    i = 0
    while collection[i:(i + size)]:
        yield collection[i:(i + size)]
        i += size


def format_identification(spectrum_solution: SpectrumSolutionSet) -> str:
    return "%s:%0.3f:(%0.3f) ->\n%s" % (
        spectrum_solution.scan.id,
        spectrum_solution.scan.precursor_information.neutral_mass,
        spectrum_solution.best_solution().score,
        str(spectrum_solution.best_solution().target))


def format_identification_batch(group: Iterable[SpectrumSolutionSet], n: int) -> str:
    representers = dict()
    group = sorted(group, key=lambda x: x.score, reverse=True)
    for ident in group:
        key = str(ident.best_solution().target)
        if key in representers:
            continue
        else:
            representers[key] = ident
    to_represent = sorted(
        representers.values(), key=lambda x: x.score, reverse=True)
    return '\n'.join(map(format_identification, to_represent[:n]))


class SearchEngineBase(TaskBase):

    def search(self, precursor_error_tolerance: float=1e-5,
               simplify: bool=True,
               batch_size: int=500,
               limit: Optional[int]=None, *args, **kwargs) -> TargetDecoySet:
        raise NotImplementedError()

    def estimate_fdr(self, *args, **kwargs) -> TargetDecoySet:
        raise NotImplementedError()



import logging
import operator

from typing import (
    Callable, DefaultDict, Dict,
    List, Mapping, Optional, Set,
    TYPE_CHECKING, Tuple, Union,
    OrderedDict, Iterable, NamedTuple)
from array import array
from dataclasses import dataclass

from matplotlib import pyplot as plt

import numpy as np

from glypy.structure.glycan_composition import HashableGlycanComposition, FrozenMonosaccharideResidue
from glycresoft.chromatogram_tree import mass_shift

from glycresoft.chromatogram_tree.mass_shift import Unmodified, Ammonium, Deoxy
from glycresoft.scoring.elution_time_grouping.structure import GlycopeptideChromatogramProxy
from glycresoft.structure.structure_loader import GlycanCompositionDeltaCache

from glycresoft._c.composition_network.graph import CachingGlycanCompositionVectorContext, GlycanCompositionVector

from glycresoft.task import LoggingMixin

logger = logging.getLogger(__name__)
logger.addHandler(logging.NullHandler())


if TYPE_CHECKING:
    from glycresoft.scoring.elution_time_grouping.model import ElutionTimeFitter
    from glycresoft.tandem.chromatogram_mapping import SpectrumMatchUpdater, TandemAnnotatedChromatogram
    from glycresoft.tandem.target_decoy import NearestValueLookUp, TargetDecoyAnalyzer
    from glycresoft.tandem.spectrum_match import SpectrumMatch, MultiScoreSpectrumMatch


def _invert_fn(fn):
    def wrapper(*args, **kwargs):
        value = fn(*args, **kwargs)
        return not value
    return wrapper


class RevisionRule(object):
    delta_glycan: HashableGlycanComposition
    mass_shift_rule: Optional['MassShiftRule']
    priority: int
    name: Optional[str]
    delta_cache: Optional[GlycanCompositionDeltaCache]

    def __init__(self, delta_glycan, mass_shift_rule=None, priority=0, name=None):
        self.delta_glycan = HashableGlycanComposition.parse(delta_glycan)
        self.mass_shift_rule = mass_shift_rule
        self.priority = priority
        self.name = name
        self.delta_cache = None

    def monosaccharides(self) -> Set[FrozenMonosaccharideResidue]:
        return set(self.delta_glycan)

    def clone(self):
        return self.__class__(
            self.delta_glycan.clone(),
            self.mass_shift_rule.clone() if self.mass_shift_rule else None,
            self.priority,
            self.name)

    def is_valid_revision(self, record: GlycopeptideChromatogramProxy,
                          new_record: GlycopeptideChromatogramProxy) -> bool:
        valid = not any(v < 0 for v in new_record.glycan_composition.values())
        if valid:
            if self.mass_shift_rule:
                return self.mass_shift_rule.valid(record)
        return valid

    def valid(self, record: GlycopeptideChromatogramProxy) -> bool:
        new_record = self(record)
        valid = self.is_valid_revision(record, new_record)
        return valid

    def with_cache(self):
        new = self.clone()
        new.delta_cache = GlycanCompositionDeltaCache(op=operator.add)
        return new

    def without_cache(self):
        return self.clone()

    def _apply(self, record: 'GlycopeptideChromatogramProxy') -> 'GlycopeptideChromatogramProxy':
        if self.delta_cache:
            new_gc = self.delta_cache(record.glycan_composition, self.delta_glycan)
            new_record = record.copy()
            new_record.update_glycan_composition(new_gc)
            return new_record
        else:
            # Re-compute the new glycan every time
            return record.shift_glycan_composition(self.delta_glycan)

    def __call__(self, record):
        result = self._apply(record)
        if self.mass_shift_rule is not None:
            return self.mass_shift_rule(result)
        return result

    def revert(self, record):
        return record.shift_glycan_composition(-self.delta_glycan)

    def invert_rule(self):
        return self.__class__(
            -self.delta_glycan, self.mass_shift_rule.invert() if self.mass_shift_rule else None, self.priority)

    def __eq__(self, other):
        try:
            return self.delta_glycan == other.delta_glycan and self.mass_shift_rule == other.mass_shift_rule
        except AttributeError:
            return self.delta_glycan == other

    def __ne__(self, other):
        return not self == other

    def __hash__(self):
        return hash(self.delta_glycan)

    def __repr__(self):
        if self.mass_shift_rule:
            template = "{self.__class__.__name__}({self.delta_glycan}, {self.mass_shift_rule}, {self.name})"
        else:
            template = "{self.__class__.__name__}({self.delta_glycan}, {self.name})"
        return template.format(self=self)


class ValidatingRevisionRule(RevisionRule):
    def __init__(self, delta_glycan, validator, mass_shift_rule=None, priority=0, name=None):
        super(ValidatingRevisionRule, self).__init__(
            delta_glycan, mass_shift_rule=mass_shift_rule, priority=priority, name=name)
        self.validator = validator

    def clone(self):
        return self.__class__(
            self.delta_glycan.clone(), self.validator, self.mass_shift_rule.clone() if self.mass_shift_rule else None,
            self.priority, self.name)

    def is_valid_revision(self, record: GlycopeptideChromatogramProxy,
                          new_record: GlycopeptideChromatogramProxy) -> bool:
        if super().is_valid_revision(record, new_record):
            return self.validator(record)
        return False

    def invert_rule(self, validator: Optional[Callable[['GlycopeptideChromatogramProxy'], bool]]=None):
        if validator is None:
            validator = _invert_fn(self.validator)
        return self.__class__(
            -self.delta_glycan, validator, self.mass_shift_rule.invert() if self.mass_shift_rule else None,
            self.priority)


class MassShiftRule(object):
    def __init__(self, mass_shift, multiplicity):
        self.mass_shift = mass_shift
        self.sign = int(multiplicity / abs(multiplicity))
        self.multiplicity = abs(multiplicity)
        self.single = abs(multiplicity) == 1

    def invert_rule(self):
        return self.__class__(self.mass_shift, -self.sign * self.multiplicity)

    def clone(self):
        return self.__class__(self.mass_shift, self.multiplicity * self.sign)

    def valid(self, record):
        assert isinstance(self.mass_shift, mass_shift.MassShift)
        if self.sign < 0:
            # Can only lose a mass shift if all the mass shifts of the analyte are able to lose
            # self.mass_shift without going negative
            for shift in record.mass_shifts:
                # The current mass shift is a compound mass shift, potentially having multiple copies of self.mass_shift
                if isinstance(shift, mass_shift.CompoundMassShift) and \
                        shift.counts.get(self.mass_shift, 0) < self.multiplicity:
                    return False
                # The current mass shift is a single simple mass shift and it isn't a match for self.mass_shift
                # or self.multiplicity > 1
                elif isinstance(shift, mass_shift.MassShift) and (shift != self.mass_shift) or \
                        (shift == self.mass_shift and not self.single):
                    return False
            return True
        else:
            # Can always gain a mass shift
            return True

    def apply(self, record):
        new = record.copy()
        new.mass_shifts = [
            m + (self.mass_shift * self.sign * self.multiplicity) for m in new.mass_shifts]
        return new

    def __call__(self, record):
        return self.apply(record)

    def __eq__(self, other):
        if other is None:
            return False
        return self.mass_shift == other.mass_shift and self.multiplicity == other.multiplicity

    def __ne__(self, other):
        return not self == other

    def __hash__(self):
        return hash(self.mass_shift)

    def __repr__(self):
        template = "{self.__class__.__name__}({self.mass_shift}, {self.multiplicity})"
        return template.format(self=self)


def modify_rules(rules: Iterable[RevisionRule], symbol_map: Mapping[str, str]):
    rules = list(rules)
    for sym_from, sym_to in symbol_map.items():
        editted: List[RevisionRule] = []
        for rule in rules:
            if sym_from in rule.delta_glycan:
                if rule.delta_cache is not None:
                    rule = rule.with_cache()
                else:
                    rule = rule.clone()
                count = rule.delta_glycan.pop(sym_from)
                rule.delta_glycan[sym_to] = count
            editted.append(rule)
        rules = editted
    return rules


def always(x):
    return True


SpectrumMatchType = Union['SpectrumMatch', 'MultiScoreSpectrumMatch']

class RevisionQueryResult(NamedTuple):
    spectrum_match: Optional[SpectrumMatchType]
    found: bool


class OriginalRevisionQueryResult(NamedTuple):
    original_spectrum_match: Optional[SpectrumMatchType]
    revision_spectrum_match: Optional[SpectrumMatchType]
    skip: bool


class RevisionValidatorBase(LoggingMixin):
    spectrum_match_builder: 'SpectrumMatchUpdater'

    def __init__(self, spectrum_match_builder, threshold_fn=always):
        self.spectrum_match_builder = spectrum_match_builder
        self.threshold_fn = threshold_fn

    def find_revision_gpsm(self, chromatogram: 'TandemAnnotatedChromatogram', revised) -> RevisionQueryResult:
        found_revision = False
        revised_gpsm = None
        try:
            revised_gpsm = chromatogram.best_match_for(revised.structure)
            found_revision = True
        except KeyError:
            if self.spectrum_match_builder is not None:
                self.spectrum_match_builder.get_spectrum_solution_sets(
                    revised,
                    chromatogram,
                    invalidate_reference=False
                )
                try:
                    revised_gpsm = chromatogram.best_match_for(revised.structure)
                    found_revision = True
                except KeyError:
                    logger.debug(
                        "Failed to find revised spectrum match for %s in %r",
                        revised.structure,
                        chromatogram)
        return RevisionQueryResult(revised_gpsm, found_revision)

    def find_gpsms(self, source: 'TandemAnnotatedChromatogram',
                   revised: GlycopeptideChromatogramProxy,
                   original: GlycopeptideChromatogramProxy) -> OriginalRevisionQueryResult:
        '''Find spectrum matches in ``source`` for the pair of alternative interpretations
        for this feature.

        Parameters
        ----------
        source : :class:`~.TandemAnnotatedChromatogram`
            The chromatogram with MS2 spectra to find spectrum matches from
        revised : :class:`~.GlycopeptideChromatogramProxy`
            The revised structure
        original : :class:`~.GlycopeptideChromatogramProxy`
            The original structure

        Returns
        -------
        original_gpsm : :class:`~.SpectrumMatchType`
            The best spectrum match for the original structure
        revised_gpsm : :class:`~.SpectrumMatchType`
            The best spectrum match for the revised structure
        skip : :class:`bool`
            Whether there were any failures to find a revised solution, telling
            the caller to skip over this one.
        '''
        original_gpsm = None
        revised_gpsm = None
        skip = False
        if source is None:
            # Can't validate without a source to read the spectrum match metadata
            skip = True
            return OriginalRevisionQueryResult(original_gpsm, revised_gpsm, skip)

        revised_gpsm, found_revision = self.find_revision_gpsm(source, revised)

        if not found_revision:
            # Can't find a spectrum match to the revised form, assume we're allowed to
            # revise.
            self.debug(
                "...... Permitting revision for %s (%0.3f) from %s to %s because revision not evaluated",
                revised.tag, revised.apex_time, original.glycan_composition, revised.glycan_composition)
            skip = True
            return OriginalRevisionQueryResult(original_gpsm, revised_gpsm, skip)

        try:
            original_gpsm = source.best_match_for(original.structure)
        except KeyError:
            # Can't find a spectrum match to the original, assume we're allowed to
            # revise.
            self.debug(
                "...... Permitting revision for %s (%0.3f) from %s to %s because original not evaluated" %
                (revised.tag, revised.apex_time, original.glycan_composition, revised.glycan_composition))
            skip = True
            return OriginalRevisionQueryResult(original_gpsm, revised_gpsm, skip)
        skip = revised_gpsm is None or original_gpsm is None
        return OriginalRevisionQueryResult(original_gpsm, revised_gpsm, skip)

    def validate(self, revised: GlycopeptideChromatogramProxy, original: GlycopeptideChromatogramProxy) -> bool:
        raise NotImplementedError()

    def __call__(self, revised: GlycopeptideChromatogramProxy, original: GlycopeptideChromatogramProxy) -> bool:
        return self.validate(revised, original)

    def msn_score_for(self, source: 'TandemAnnotatedChromatogram', solution: GlycopeptideChromatogramProxy) -> float:
        gpsm, found = self.find_revision_gpsm(source, solution)
        if not found or gpsm is None:
            return 0
        return gpsm.score


class PeptideYUtilizationPreservingRevisionValidator(RevisionValidatorBase):
    '''Prevent a revision that leads to substantial loss of peptide+Y ion-explainable
    signal.

    This does not use peptide backbone fragments because they are by definition the
    revision algorithm only updates the glycan composition, not the peptide backbone.
    '''
    threshold: float

    def __init__(self, threshold=0.85, spectrum_match_builder=None, threshold_fn=always):
        super(PeptideYUtilizationPreservingRevisionValidator, self).__init__(
            spectrum_match_builder, threshold_fn)
        self.threshold = threshold

    def validate(self, revised, original):
        source = revised.source
        if source is None:
            # Can't validate without a source to read the spectrum match metadata
            return True

        original_gpsm, revised_gpsm, skip = self.find_gpsms(
            source, revised, original)
        if skip:
            return True
        if hasattr(original_gpsm, 'score_set'):
            original_utilization = original_gpsm.score_set.stub_glycopeptide_intensity_utilization
            if not original_utilization:
                # Anything is better than or equal to zero
                return True

            revised_utilization = revised_gpsm.score_set.stub_glycopeptide_intensity_utilization
            utilization_ratio = revised_utilization / original_utilization
            valid = utilization_ratio > self.threshold

            self.debug(
                "...... Checking revision by peptide+Y ions for %s (%0.3f) from %s to %s: %0.3f / %0.3f: %r" %
                (revised.tag, revised.apex_time, original.glycan_composition, revised.glycan_composition,
                revised_utilization, original_utilization, valid))
        else:
            # TODO: Make this check explicit by directly calculating the metric here.
            valid = True
        return valid


class OxoniumIonRequiringUtilizationRevisionValidator(RevisionValidatorBase):
    scale: float

    def __init__(self, scale=1.0, spectrum_match_builder=None, threshold_fn=always):
        super(OxoniumIonRequiringUtilizationRevisionValidator,
              self).__init__(spectrum_match_builder, threshold_fn)
        self.scale = scale

    def validate(self, revised: GlycopeptideChromatogramProxy, original: GlycopeptideChromatogramProxy) -> bool:
        source = revised.source
        if source is None:
            # Can't validate without a source to read the spectrum match metadata
            return True

        original_gpsm, revised_gpsm, skip = self.find_gpsms(
            source, revised, original)
        if skip:
            return True
        if hasattr(original_gpsm, 'score_set'):
            original_utilization: float = original_gpsm.score_set.oxonium_ion_intensity_utilization
            revised_utilization: float = revised_gpsm.score_set.oxonium_ion_intensity_utilization

            threshold = original_utilization * self.scale
            valid = (revised_utilization - threshold) >= 0
            self.debug(
                "...... Checking revision by oxonium ions for %s (%0.3f) from %s to %s: %0.3f / %0.3f: %r" %
                (revised.tag, revised.apex_time, original.glycan_composition, revised.glycan_composition,
                revised_utilization, original_utilization, valid))
        else:
            # TODO: Make this check explicit by directly calculating the metric here.
            valid = True
        return valid


class CompoundRevisionValidator(object):
    validators: List[RevisionValidatorBase]

    def __init__(self, validators=None):
        if validators is None:
            validators = []
        self.validators = list(validators)

    @property
    def spectrum_match_builder(self):
        return self.validators[0].spectrum_match_builder

    def validate(self, revised, original):
        for rule in self.validators:
            if not rule(revised, original):
                return False
        return True

    def __call__(self, revised, original):
        return self.validate(revised, original)

    def msn_score_for(self, source, solution: GlycopeptideChromatogramProxy) -> float:
        return self.validators[0].msn_score_for(source, solution)


def _new_array():
    return array('d')


def display_table(rows: List[str], headers: List[str]) -> str:
    rows = [headers] + rows
    widths = list(map(lambda col: max(map(len, col)), zip(*rows)))

    buffer = []
    for i, row in enumerate(rows):
        buffer.append(' | '.join(col.center(w) for col, w in zip(row, widths)))
        if i == 0:
            buffer.append(
                '|'.join('-' * (w + 2 if j > 0 else w + 1)
                         for j, (col, w) in enumerate(zip(row, widths)))
            )

    return '\n'.join(buffer)


@dataclass
class RevisionEvent:
    rt_score: float
    chromatogram_record: GlycopeptideChromatogramProxy
    rule: Optional[RevisionRule]
    msn_score: float


class ValidatedGlycome:
    valid_glycans: Set[HashableGlycanComposition]
    monosaccharides: Set[FrozenMonosaccharideResidue]

    def __init__(self, valid_glycans):
        self.valid_glycans = valid_glycans
        self.monosaccharides = self._collect_monosaccharides()

    def _collect_monosaccharides(self):
        keys = set()
        for gc in self.valid_glycans:
            keys.update(gc)
        return keys

    def __contains__(self, glycan_composition: HashableGlycanComposition) -> bool:
        return glycan_composition in self.valid_glycans

    def __bool__(self):
        return bool(self.valid_glycans)

    def __nonzero__(self):
        return bool(self.valid_glycans)

    def __iter__(self):
        return iter(self.valid_glycans)

    def __len__(self):
        return len(self.valid_glycans)

    def check(self, record: GlycopeptideChromatogramProxy) -> bool:
        return record.glycan_composition in self.valid_glycans


class IndexedValidatedGlycome(ValidatedGlycome):
    indexer: CachingGlycanCompositionVectorContext
    indexed_valid_glycans: Set[GlycanCompositionVector]

    def __init__(self, valid_glycans):
        super().__init__(valid_glycans)
        self.indexer = CachingGlycanCompositionVectorContext(self.monosaccharides)
        self.indexed_valid_glycans = {
            self.indexer.encode(gc) for gc in self.valid_glycans
        }

    def encode(self, record: GlycopeptideChromatogramProxy):
        if "_indexed_glycan_composition" in record.kwargs:
            return record.kwargs["_indexed_glycan_composition"]
        vec = self.indexer.encode(record.glycan_composition)
        record.kwargs["_indexed_glycan_composition"] = vec
        return vec

    def check(self, record: GlycopeptideChromatogramProxy) -> bool:
        return self.encode(record) in self.indexed_valid_glycans



class ModelReviser(LoggingMixin):
    valid_glycans: Optional[ValidatedGlycome]
    revision_validator: Optional[RevisionValidatorBase]

    model: 'ElutionTimeFitter'
    rules: 'RevisionRuleList'
    chromatograms: List[GlycopeptideChromatogramProxy]

    original_scores: array
    original_times: array

    alternative_records: DefaultDict[RevisionRule, List[GlycopeptideChromatogramProxy]]
    alternative_scores: DefaultDict[RevisionRule, array]
    alternative_times: DefaultDict[RevisionRule, array]

    def __init__(self, model, rules, chromatograms=None, valid_glycans=None, revision_validator=None):
        if chromatograms is None:
            chromatograms = model.chromatograms
        self.model = model
        self.chromatograms = chromatograms
        self.original_scores = array('d')
        self.original_times = array('d')
        self.rules = list(rules)
        self.alternative_records = DefaultDict(list)
        self.alternative_scores = DefaultDict(_new_array)
        self.alternative_times = DefaultDict(_new_array)
        if valid_glycans is not None:
            if isinstance(valid_glycans, set):
                valid_glycans = ValidatedGlycome(valid_glycans)
        self.valid_glycans = valid_glycans
        self.revision_validator = revision_validator

    def rescore(self, case):
        return self.model.score(case)

    def modify_rules(self, symbol_map):
        self.rules = self.rules.modify_rules(symbol_map)
        return self

    def propose_revisions(self, case):
        propositions = {
            None: (case, self.rescore(case)),
        }
        for rule in self.rules:
            if rule.valid(case):
                rec = rule(case)
                if self.valid_glycans and not self.valid_glycans.check(rec):
                    continue
                propositions[rule] = (rec, self.rescore(rec))
        return propositions

    def process_rule(self, rule: RevisionRule):
        alts = []
        alt_scores = array('d')
        alt_times = array('d')
        for case in self.chromatograms:
            rec = rule(case)
            if rule.is_valid_revision(case, rec):
                if self.valid_glycans and not self.valid_glycans.check(rec):
                    alts.append(None)
                    alt_scores.append(0.0)
                    alt_times.append(0.0)
                    continue
                alts.append(rec)
                alt_times.append(self.model.predict(rec))
                alt_scores.append(self.rescore(rec))
            else:
                alts.append(None)
                alt_scores.append(0.0)
                alt_times.append(0.0)
        self.alternative_records[rule] = alts
        self.alternative_scores[rule] = alt_scores
        self.alternative_times[rule] = alt_times

    def process_model(self):
        scores = array('d')
        times = array('d')
        for case in self.chromatograms:
            scores.append(self.rescore(case))
            times.append(self.model.predict(case))
        self.original_scores = scores
        self.original_times = times

    def evaluate(self):
        self.process_model()
        for rule in self.rules:
            self.process_rule(rule)

    def select_revision_atlernative(self, alternatives: List[RevisionEvent],
                                    original: GlycopeptideChromatogramProxy,
                                    minimum_rt_score: float) -> GlycopeptideChromatogramProxy:
        alternatives = [alt for alt in alternatives if alt.rt_score > minimum_rt_score]
        alternatives.sort(key=lambda x: x.msn_score, reverse=True)

        table = [
            (str(alt.chromatogram_record.glycan_composition),
            (','.join(map(lambda x: x.name, alt.chromatogram_record.mass_shifts))),
            (alt.rule.name if alt.rule is not None else '-'),
             '%0.3f' % alt.rt_score, '%0.3f' % alt.msn_score)
            for alt in alternatives
        ]

        if len({alt.chromatogram_record.glycan_composition for alt in alternatives}) > 1:
            self.log('\n' + display_table(table, ['Glycan Composition', 'Mass Shifts',
                                                  'Revision Rule', 'RT Score', 'MSn Score']))

        best_event = alternatives[0]
        best_record = best_event.chromatogram_record
        best_rule = best_event.rule
        if best_rule is not None: # If *somehow* the best solution at the time this is called is still unchanged?
            best_record.revised_from = (best_rule, original)
        return best_record

    def _msn_score_if_available(self, source, solution: GlycopeptideChromatogramProxy) -> float:
        if self.revision_validator is None:
            return 0.0
        return self.revision_validator.msn_score_for(source, solution)

    def revise(self, threshold=0.2, delta_threshold=0.2, minimum_time_difference=0.5):
        chromatograms = self.chromatograms
        original_scores = self.original_scores
        original_times = self.original_times
        next_round = []
        for i in range(len(chromatograms)):
            original_score = best_score = original_scores[i]
            best_rule = None
            delta_best_score = float('inf')
            original_solution = best_record = chromatograms[i]
            original_time = original_times[i]

            alternative_solutions = [
                RevisionEvent(original_score, original_solution, None,
                              self._msn_score_if_available(
                                original_solution.source,
                                original_solution)
                )
            ]

            for rule in self.rules:
                alt_rec = self.alternative_records[rule][i]
                alt_score = self.alternative_scores[rule][i]
                alt_time = self.alternative_times[rule][i]
                if not np.isclose(alt_score, 0.0) and abs(alt_time - original_time) > minimum_time_difference:
                    if self.revision_validator and not self.revision_validator(alt_rec, original_solution):
                        continue
                    alternative_solutions.append(
                        RevisionEvent(
                            alt_score, alt_rec, rule,
                            self._msn_score_if_available(
                                original_solution.source, alt_rec)))
                    if alt_score > best_score:
                        delta_best_score = alt_score - original_score
                        best_score = alt_score
                        best_rule = rule
                        best_record = alt_rec


            if best_score > threshold and delta_best_score > delta_threshold:
                if best_rule is not None:
                    best_record = self.select_revision_atlernative(
                        alternative_solutions, original_solution, best_score * 0.9)
                next_round.append(best_record)
            else:
                next_round.append(chromatograms[i])
        return next_round


dhex = FrozenMonosaccharideResidue.from_iupac_lite("d-Hex")
fuc = FrozenMonosaccharideResidue.from_iupac_lite("Fuc")
neuac = FrozenMonosaccharideResidue.from_iupac_lite("Neu5Ac")
hexnac = FrozenMonosaccharideResidue.from_iupac_lite("HexNAc")

# The correct glycan was incorrectly assigned as the unadducted form of
# another glycan.
#
# Gain an ammonium adduct and a NeuAc, lose a Hex and Fuc
AmmoniumMaskedRule = RevisionRule(
    HashableGlycanComposition(Hex=-1, Fuc=-1, Neu5Ac=1),
    mass_shift_rule=MassShiftRule(Ammonium, 1), priority=1, name="Ammonium Masked")

AmmoniumMaskedNeuGcRule = RevisionRule(
    HashableGlycanComposition(Hex=-2, Neu5Gc=1),
    mass_shift_rule=MassShiftRule(Ammonium, 1), priority=1, name="Ammonium Masked NeuGc")

# The correct glycan was incorrectly assigned as the ammonium adducted form of another
# glycan
AmmoniumUnmaskedRule = RevisionRule(
    HashableGlycanComposition(Hex=1, Fuc=1, Neu5Ac=-1),
    mass_shift_rule=MassShiftRule(Ammonium, -1), priority=1, name="Ammonium Unmasked")

AmmoniumUnmaskedNeuGcRule = RevisionRule(
    HashableGlycanComposition(Hex=2, Neu5Gc=-1),
    mass_shift_rule=MassShiftRule(Ammonium, -1), priority=1, name="Ammonium Unmasked NeuGc")

# The glycan was incorrectly assigned identified because of an error in monoisotopic peak
# assignment.
IsotopeRule = RevisionRule(HashableGlycanComposition(Fuc=-2, Neu5Ac=1), name="Isotope Error")
IsotopeRule2 = RevisionRule(HashableGlycanComposition(Fuc=-4, Neu5Ac=2), name="Isotope Error 2")

IsotopeRuleNeuGc = RevisionRule(HashableGlycanComposition(Fuc=-1, Hex=-1, NeuGc=1), name="Isotope Error NeuGc")

HexNAc2NeuAc2ToHex6Deoxy = ValidatingRevisionRule(
    HashableGlycanComposition(Hex=-6, HexNAc=2, Neu5Ac=2),
    mass_shift_rule=MassShiftRule(Deoxy, 1),
    validator=lambda x: x.glycan_composition[neuac] == 0 and x.glycan_composition.query(dhex) == 0 and \
            x.glycan_composition[hexnac] == 2 and x.mass_shifts == [Unmodified],
    name="Rare precursor deoxidation followed by large mass ambiguity")

HexNAc2NeuAc2ToHex6AmmoniumRule = ValidatingRevisionRule(
    HashableGlycanComposition(Hex=-6, HexNAc=2, Neu5Ac=2),
    mass_shift_rule=MassShiftRule(Ammonium, 1),
    validator=lambda x: x.glycan_composition[neuac] == 0 and x.glycan_composition.query(dhex) == 0 and \
            x.glycan_composition[hexnac] == 2,
    name="Ammonium Masked followed by Large Mass Ambiguity")


# Not yet in use, needs to be explored more. In practice, this cannot even be used
# with such a small chromatographic shift (<= 0.5 minutes) on fucose and hexose.
IsotopeAmmoniumFucToHex = RevisionRule(HashableGlycanComposition(Fuc=-1, Hex=1),
                                       mass_shift_rule=MassShiftRule(Ammonium, 1),
                                       name='Ammonium Isotope Error')

# Not yet in use, needs to be explored more
dHex2HexNAc2NeuAc1ToHex6AmmoniumRule = ValidatingRevisionRule(
    HashableGlycanComposition(Fuc=2, Hex=-6, HexNAc=2, Neu5Ac=1),
    mass_shift_rule=MassShiftRule(Ammonium, 1),
    validator=lambda x: x.glycan_composition[neuac] == 0 and x.glycan_composition.query(
        dhex) == 0 and x.glycan_composition[hexnac] == 1,
    name="Ammonium Masked followed by Large Mass Ambiguity II")


HexNAc2Fuc1NeuAc2ToHex7 = ValidatingRevisionRule(
    HashableGlycanComposition(Hex=-7, HexNAc=2, Neu5Ac=2, Fuc=1),
    validator=lambda x: x.glycan_composition[neuac] == 0 and x.glycan_composition.query(dhex) == 0 and \
            x.glycan_composition[hexnac] == 2,
    name="Large Mass Ambiguity")

# Moving a hydroxyl group from another monosaccharide onto a Sialic acid or vis-versa obscures their mass identity
NeuGc1Fuc1ToNeuAc1Hex1Rule = RevisionRule(
    HashableGlycanComposition(Neu5Ac=1, Hex=1, Neu5Gc=-1, Fuc=-1), name="NeuAc Masked By NeuGc")
NeuAc1Hex1ToNeuGc1Fuc1Rule = RevisionRule(
    HashableGlycanComposition(Neu5Ac=-1, Hex=-1, Neu5Gc=1, Fuc=1), name="NeuGc Masked By NeuAc")

Sulfate1HexNAc2ToHex3Rule = RevisionRule(
    HashableGlycanComposition(HexNAc=2, sulfate=1, Hex=-3), name="Sulfate + 2 HexNAc Masked By 3 Hex")
Hex3ToSulfate1HexNAc2Rule = Sulfate1HexNAc2ToHex3Rule.invert_rule()
Hex3ToSulfate1HexNAc2Rule.name = "3 Hex Masked By Sulfate + 2 HexNAc"

Phosphate1HexNAc2ToHex3Rule = RevisionRule(
    HashableGlycanComposition(HexNAc=2, phosphate=1, Hex=-3), name="Phosphate + 2 HexNAc Masked By 3 Hex")
Hex3ToPhosphate1HexNAc2Rule = Phosphate1HexNAc2ToHex3Rule.invert_rule()
Hex3ToPhosphate1HexNAc2Rule.name = "3 Hex Masked By Phosphate + 2 HexNAc"

SulfateToPhosphateRule = RevisionRule(
    HashableGlycanComposition(sulfate=-1, phosphate=1),
    name="Phosphate Masked By Sulfate")
PhosphateToSulfateRule = SulfateToPhosphateRule.invert_rule()
PhosphateToSulfateRule.name = "Sulfate Masked By Phosphate"


class RevisionRuleList(object):
    rules: List[RevisionRule]
    by_name: Dict[str, RevisionRule]

    def __init__(self, rules):
        self.rules = list(rules)
        self.by_name = {
            rule.name: rule for rule in self.rules
            if rule.name
        }

    def modify_rules(self, symbol_map):
        self.rules = modify_rules(self.rules, symbol_map)
        return self

    def with_cache(self):
        return self.__class__([
            rule.with_cache() for rule in self.rules
        ])

    def filter_defined(self, valid_glycans: ValidatedGlycome):
        acc = []
        for rule in self.rules:
            keys = rule.monosaccharides()
            if keys == (keys & valid_glycans.monosaccharides):
                acc.append(rule)
        self.rules = acc
        return self

    def __getitem__(self, i):
        if i in self.by_name:
            return self.by_name[i]
        return self.rules[i]

    def __len__(self):
        return len(self.rules)

    def __iter__(self):
        return iter(self.rules)

    def __repr__(self):
        return "{self.__class__.__name__}({self.rules})".format(self=self)


class IntervalModelReviser(ModelReviser):
    alpha: float = 0.01

    def rescore(self, case):
        return self.model.score_interval(case, alpha=self.alpha)


# In reviser, combine the score from the absolute coordinate with the
# score relative to each other glycoform in the group? This might look
# something like the old recalibrator code, but adapted for the interval_score
# method


class FDREstimatorBase(object):
    estimator: 'TargetDecoyAnalyzer'

    def estimate_fdr(self):
        from glycresoft.tandem.target_decoy import TargetDecoyAnalyzer
        self.estimator = TargetDecoyAnalyzer(
            self.target_scores.view([('score', np.float64)]).view(np.recarray),
            self.decoy_scores[self.decoy_is_valid].view([('score', np.float64)]).view(np.recarray))

    @property
    def q_value_map(self):
        return self.estimator.q_value_map

    def score_for_fdr(self, fdr_value: float) -> float:
        return self.estimator.score_for_fdr(fdr_value)

    def plot(self):
        ax = self.estimator.plot()
        ax.set_xlim(-0.01, 1)
        return ax


class RuleBasedFDREstimator(FDREstimatorBase):
    rule: RevisionRule

    chromatograms: Union[List[GlycopeptideChromatogramProxy], np.ndarray]
    decoy_chromatograms: Union[List[GlycopeptideChromatogramProxy], np.ndarray]

    target_scores: np.ndarray
    decoy_scores: np.ndarray

    target_predictions: np.ndarray
    decoy_predictions: np.ndarray
    decoy_residuals: np.ndarray

    decoy_is_valid: np.ndarray
    over_time: OrderedDict

    def __init__(self, rule, chromatograms, rt_model, valid_glycans=None):
        self.rule = rule
        self.chromatograms = chromatograms
        self.decoy_chromatograms = []
        self.rt_model = rt_model
        self.target_scores = np.array([])
        self.decoy_scores = np.array([])
        self.decoy_is_valid = np.array([])
        self.valid_glycans = valid_glycans
        self.estimator = None
        self.over_time = OrderedDict()

        self.prepare()

    @property
    def name(self):
        return self.rule.name

    def prepare(self):
        self.target_scores = np.array([self.rt_model.score_interval(p, 0.01) for p in self.chromatograms])
        self.target_times = np.array([p.apex_time for p in self.chromatograms])
        self.target_predictions = np.array([
            self.rt_model.predict(p) for p in self.chromatograms
        ])
        self.target_residuals = self.target_times - self.target_predictions
        decoy_scores = []
        decoy_is_valid = []
        self.decoy_chromatograms = []
        for p in self.chromatograms:
            p = self.rule(p)
            self.decoy_chromatograms.append(p)
            decoy_is_valid.append(p.glycan_composition in self.valid_glycans if self.valid_glycans else True)
            if self.rule.valid(p):
                decoy_scores.append(self.rt_model.score_interval(p, 0.01))
            else:
                decoy_scores.append(np.nan)

        self.decoy_times = np.array([p.apex_time for p in self.decoy_chromatograms])
        self.decoy_predictions = np.array([
            self.rt_model.predict(p) for p in self.decoy_chromatograms
        ])
        self.decoy_residuals = self.decoy_times - self.decoy_predictions


        self.decoy_scores = np.array(decoy_scores)
        self.decoy_is_valid = np.array(decoy_is_valid)

        self.target_scores[np.isnan(self.target_scores)] = 0.0
        self.decoy_scores[np.isnan(self.decoy_scores)] = 0.0
        self.chromatograms = np.array(self.chromatograms)
        self.decoy_chromatograms = np.array(self.decoy_chromatograms)

        self.estimate_fdr()

    def __repr__(self):
        template = "{self.__class__.__name__}({self.rule})"
        return template.format(self=self)

    def get_residuals_from_interval(self, span):
        target_mask = [span.contains(i) for i in self.target_times]
        target_residuals = self.target_residuals[target_mask]
        target_residuals = target_residuals[~np.isnan(target_residuals)]

        decoy_mask = [span.contains(i) for i in self.decoy_times]
        decoy_residuals = self.decoy_residuals[decoy_mask & self.decoy_is_valid]
        decoy_residuals = decoy_residuals[~np.isnan(decoy_residuals)]
        return target_residuals, decoy_residuals

    def get_scores_from_interval(self, span):
        target_mask = [span.contains(i) for i in self.target_times]
        target_scores = self.target_scores[target_mask]
        target_scores = target_scores[~np.isnan(target_scores)]

        decoy_mask = [span.contains(i) for i in self.decoy_times]
        decoy_scores = self.decoy_scores[decoy_mask &
                                               self.decoy_is_valid]
        decoy_scores = decoy_scores[~np.isnan(decoy_scores)]
        return target_scores, decoy_scores

    def get_scores_from_mask(self, mask):
        target_scores = self.target_scores[mask]
        target_scores = target_scores[~np.isnan(target_scores)]

        decoy_scores = self.decoy_scores[mask &
                                         self.decoy_is_valid]
        decoy_scores = decoy_scores[~np.isnan(decoy_scores)]
        return target_scores, decoy_scores

    def get_interval_masks(self):
        spans = [mod for mod in self.rt_model.models.values()]
        centers = np.array([s.centroid for s in spans])
        masks = [
            np.fromiter(map(span.contains, self.target_times), bool) for span in spans]
        return masks, centers

    def fit_over_time(self):
        masks, centers = self.get_interval_masks()
        for i, mask in enumerate(masks):
            target_scores, decoy_scores = self.get_scores_from_mask(mask)
            facet = FDRFacet(target_scores, decoy_scores)
            self.over_time[centers[i]] = facet

    def __getstate__(self):
        state = {
            "estimator": self.estimator,
            "target_scores": self.target_scores,
            "target_predictions": self.target_predictions,
            "target_times": self.target_times,
            "decoy_scores": self.decoy_scores,
            "decoy_predictions": self.decoy_predictions,
            "decoy_times": self.decoy_times,
            "decoy_is_valid": self.decoy_is_valid,
            "rule": self.rule,
            "over_time": self.over_time
        }
        return state

    def __setstate__(self, state):
        self.estimator = state['estimator']
        self.target_scores = state['target_scores']
        self.decoy_scores = state['decoy_scores']
        self.decoy_is_valid = state['decoy_is_valid']
        self.rule = state['rule']

        self.target_times = state.get("target_times", np.array([]))
        self.target_predictions = state.get("target_predictions", np.array([]))

        self.decoy_times = state.get("decoy_times", np.array([]))
        self.decoy_predictions = state.get("decoy_predictions", np.array([]))
        self.over_time = state.get("over_time", OrderedDict())

        self.target_residuals = self.target_times - self.target_predictions
        self.decoy_residuals = self.decoy_times - self.decoy_predictions

        self.rt_model = None
        self.chromatograms = None
        self.decoy_chromatograms = None
        self.valid_glycans = None

    def copy(self):
        proto, newargs, state = self.__reduce__()
        dup = proto(*newargs)
        dup.__setstate__(state)

        dup.rt_model = self.rt_model
        dup.chromatograms = self.chromatograms
        dup.decoy_chromatograms = self.decoy_chromatograms
        dup.valid_glycans = self.valid_glycans
        return dup


class FDRFacet(FDREstimatorBase):
    def __init__(self, target_scores, decoy_scores):
        self.target_scores = target_scores
        self.decoy_scores = decoy_scores
        if decoy_scores is None:
            self.decoy_is_valid = np.array([], dtype=bool)
        else:
            self.decoy_is_valid = np.ones_like(decoy_scores).astype(bool)
        if self.target_scores is not None and self.decoy_scores is not None:
            self.estimate_fdr()

    def __getstate__(self):
        state = {
            "estimator": self.estimator
        }
        return state

    def __setstate__(self, state):
        self.estimator = state['estimator']
        self.target_scores = self.estimator.targets['score']
        self.decoy_scores = self.estimator.decoys['score']
        self.decoy_is_valid = np.ones_like(self.decoy_scores).astype(bool)

    def __reduce__(self):
        return self.__class__, (None, None), self.__getstate__()


def make_normalized_monotonic_bell(X: np.ndarray, Y: np.ndarray, symmetric: bool=False) -> np.ndarray:
    center = np.abs(X).argmin()
    Ynew = np.zeros_like(Y)
    last_y = None
    for i in range(center, X.size):
        if last_y is None:
            last_y = Y[i]
        if Y[i] > last_y:
            last_y = Y[i]
        Ynew[i] = last_y

    last_y = None
    for i in range(center, -1, -1):
        if last_y is None:
            last_y = Y[i]
        if Y[i] > last_y:
            last_y = Y[i]
        Ynew[i] = last_y

    if symmetric:
        Ynew = (Ynew + Ynew[::-1]) / 2

    try:
        maximum = Ynew[:center].max()
        Ynew[:center] /= maximum
    except ValueError:
        pass

    try:
        maximum = Ynew[center:].max()
        Ynew[center:] /= maximum
    except ValueError:
        pass

    return Ynew


def dropna(array: np.ndarray) -> np.ndarray:
    mask = ~np.isnan(array)
    return array[mask]


class ResidualFDREstimator(object):
    rules: RevisionRuleList
    residual_mapper: 'PosteriorErrorToScore'

    def __init__(self, rules, rt_model=None, residual_mapper=None):
        self.rules = rules
        self.rt_model = rt_model
        self.residual_mapper = residual_mapper
        if residual_mapper is None:
            self.fit()

    def _extract_scores_from_rules(self) -> Tuple[np.ndarray, np.ndarray]:
        rule = self.rules[0]
        target_scores = rule.target_residuals[
            ~np.isnan(rule.target_residuals)]
        all_decoys = []
        for rule in self.rules:
            all_decoys.extend(rule.decoy_residuals[
                ~np.isnan(rule.decoy_residuals) & rule.decoy_is_valid])
        return target_scores, np.array(all_decoys)

    def _extract_scores_from_model(self, rt_model, seed=1, n_resamples=1) -> Tuple[np.ndarray, np.ndarray]:
        rng = np.random.RandomState(seed)
        decoys = np.array([])
        for i in range(n_resamples):
            decoys = np.concatenate(
                (
                    decoys,
                    dropna(
                        rt_model._summary_statistics['apex_time_array'] - rng.permutation(
                            rt_model._summary_statistics['predicted_apex_time_array'].copy())
                    )
                )
            )

        targets = dropna(rt_model._summary_statistics['residuals_array'])
        return targets, decoys

    def fit(self):
        models: List[PosteriorErrorToScore] = []
        if self.rules:
            try:
                fit_from_rules = self.fit_from_rules()
                models.append(fit_from_rules)
            except Exception:
                logger.error("Failed to fit FDR from delta rules", exc_info=True)

        if self.rt_model:
            try:
                fit_from_rt_model = self.fit_from_rt_model()
                models.append(fit_from_rt_model)
            except Exception as err:
                show_traceback = str(err) != "No non-zero range in the posterior error distribution."
                logger.error(
                    "Failed to fit FDR from permutations: %r", err, exc_info=show_traceback)
        if models:
            models.sort(key=lambda x: x.width_at_half_max())
            model = models[0]
            self.residual_mapper = model
        else:
            raise ValueError("Could not fit any FDR model for retention time!")

    def fit_from_rules(self):
        from glycresoft.tandem.glycopeptide.dynamic_generation.multipart_fdr import (
            FiniteMixtureModelFDREstimatorDecoyGaussian
        )
        target_scores, all_decoy_scores = self._extract_scores_from_rules()
        fmm = FiniteMixtureModelFDREstimatorDecoyGaussian(all_decoy_scores, target_scores)
        fmm.fit(max_target_components=3,
                max_decoy_components=len(self.rules) * 2)
        return PosteriorErrorToScore.from_model(fmm)

    def fit_from_rt_model(self):
        from glycresoft.tandem.glycopeptide.dynamic_generation.multipart_fdr import (
            FiniteMixtureModelFDREstimatorDecoyGaussian
        )
        target_scores, all_decoy_scores = self._extract_scores_from_model(self.rt_model)
        fmm = FiniteMixtureModelFDREstimatorDecoyGaussian(
            all_decoy_scores, target_scores)
        fmm.fit(max_target_components=3)
        return PosteriorErrorToScore.from_model(fmm)

    def bounds_for_probability(self, probability: float) -> np.ndarray:
        return self.residual_mapper.bounds_for_probability(probability)

    def __reduce__(self):
        return self.__class__, (self.rules, None, self.residual_mapper)

    def __getitem__(self, i):
        return self.rules[i]

    def __len__(self):
        return len(self.rules)

    def __iter__(self):
        return iter(self.rules)

    def plot(self, ax=None, **kwargs):
        return self.residual_mapper.plot(ax, **kwargs)

    def score(self, chromatogram: GlycopeptideChromatogramProxy, *args, **kwargs) -> float:
        t_pred = self.rt_model.predict(chromatogram, *args, **kwargs)
        residual = chromatogram.apex_time - t_pred
        return self.residual_mapper(residual)

    def __call__(self, chromatogram: GlycopeptideChromatogramProxy, *args, **kwds) -> float:
        return self.score(chromatogram, *args, **kwds)


def _prepare_domain(target_scores, decoy_scores, delta=0.1):
    lo = min(target_scores.min(), decoy_scores.min())
    hi = max(target_scores.max(), decoy_scores.max())
    sign = np.sign(lo)
    lo = sign * max(abs(lo), abs(hi))
    sign = np.sign(hi)
    hi = sign * max(abs(lo), abs(hi))
    domain = np.concatenate(
        (np.arange(lo, 0, delta), np.arange(0.0, hi + delta, delta), ))
    return domain


class PosteriorErrorToScore(object):
    mapper: 'NearestValueLookUp'
    domain: np.ndarray
    normalized_score: np.ndarray

    @classmethod
    def from_model(cls, model, delta=0.1, symmetric=True):
        domain = _prepare_domain(model.target_scores, model.decoy_scores, delta=delta)
        return cls(model, domain, symmetric=symmetric)

    def __init__(self, model, domain, symmetric=True):
        self.model = model
        self.domain = domain
        self.normalized_score = None
        self.mapper = None
        if self.model is not None:
            self._create_normalized(symmetric=symmetric)

    def _create_normalized(self, symmetric=True):
        from glycresoft.tandem.target_decoy import NearestValueLookUp
        Y = self.model.estimate_posterior_error_probability(self.domain)
        self.normalized_score = np.clip(
            1 - make_normalized_monotonic_bell(self.domain, Y, symmetric=symmetric), 0, 1)
        support = np.where(self.normalized_score != 0)[0]
        if len(support) == 0:
            raise ValueError("No non-zero range in the posterior error distribution.")
        lo, hi = support[[0, -1]]
        lo = max(lo - 5, 0)
        hi += 5
        self.domain = self.domain[lo:hi]
        self.normalized_score = self.normalized_score[lo:hi]
        self.mapper = NearestValueLookUp(zip(self.domain, self.normalized_score))

    def __call__(self, value):
        if isinstance(value, (np.ndarray, Iterable)):
            return np.array([self.mapper[v] for v in value])
        return self.mapper[value]

    def __getstate__(self):
        return {
            "mapper": self.mapper,
        }

    def __setstate__(self, state):
        self.mapper = state["mapper"]
        self.domain, self.normalized_score = map(np.array, zip(*self.mapper.items))

    def __reduce__(self):
        return self.__class__, (None, None), self.__getstate__()

    def copy(self):
        proto, newargs, state = self.__reduce__()
        dup = proto(*newargs)
        dup.__setstate__(state)
        dup.model = self.model
        return dup

    def bounds_for_probability(self, probability: float) -> np.ndarray:
        xbounds = np.where(self.normalized_score >= probability)[0]
        if len(xbounds) == 0:
            return np.array([0.0, 0.0])
        lo, hi = xbounds[[0, -1]]
        return self.domain[[lo, hi]]

    def at_half_max(self):
        half_max = self.normalized_score.max() / 2
        return self.bounds_for_probability(half_max)

    def width_at_half_max(self) -> float:
        lo, hi = self.at_half_max()
        return hi - lo

    def plot(self, ax=None):
        if ax is None:
            _fig, ax = plt.subplots(1, 1)
        ax.plot(self.domain, self.normalized_score)
        ax.set_xlabel(r"$t - \hat{t}$")
        ax.set_ylabel("Probability")
        return ax
